import abc
from enum import Enum
from dataclasses import dataclass
from collections.abc import Sequence, Collection
from typing import Union, Literal, Optional, overload

import numpy as np

from qm._loc import _get_loc
from qm.type_hinting import NumberT
from qm.grpc.qm.pb import inc_qua_pb2
from qm.qua._dsl._utils import _declare_save
from qm.qua._dsl._type_hints import OneOrMore
from qm.exceptions import QmQuaException, SaveToAdcTraceException
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._dsl.stream_processing.stream_processing import ResultStreamSource
from qm.qua._dsl.stream_processing.direct_stream_processing_interface import DirectStreamSourceInterface
from qm.qua._expressions import (
    NSize,
    QuaIO,
    Scalar,
    StructT,
    QuaVariable,
    QuaArrayCell,
    ScalarOfAnyType,
    QuaArrayVariable,
    QuaStructReference,
    QuaArrayInputStream,
    QuaStructArrayVariable,
    QuaVariableInputStream,
    to_scalar_pb_expression,
    create_qua_scalar_expression,
)


class DeclarationType(Enum):
    EmptyScalar = 0
    InitScalar = 1
    EmptyArray = 2
    InitArray = 3


class _DeclarationParams(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def create_variable(
        self, var_name: str, t: type[NumberT]
    ) -> Union[QuaArrayVariable[NumberT], QuaVariable[NumberT]]:
        pass

    @abc.abstractmethod
    def create_input_stream(
        self, var_name: str, t: type[NumberT]
    ) -> Union[QuaArrayInputStream[NumberT], QuaVariableInputStream[NumberT]]:
        pass


@dataclass
class _ArrayDeclarationParams(_DeclarationParams):
    size: int
    values: Sequence[Union[bool, int, float]]

    def create_variable(self, var_name: str, t: type[NumberT]) -> QuaArrayVariable[NumberT]:
        return QuaArrayVariable(var_name, t, size=self.size, init_value=self.values)

    def create_input_stream(self, var_name: str, t: type[NumberT]) -> QuaArrayInputStream[NumberT]:
        return QuaArrayInputStream(var_name, t, size=self.size, init_value=self.values)


@dataclass
class _StructArrayDeclarationParams(_DeclarationParams):
    size: int
    position: int
    struct_ref: QuaStructReference

    def create_variable(self, var_name: str, t: type[NumberT]) -> QuaStructArrayVariable[NumberT, NSize]:
        return QuaStructArrayVariable(var_name, t, size=self.size, position=self.position, struct=self.struct_ref)

    def create_input_stream(self, var_name: str, t: type[NumberT]) -> QuaArrayInputStream[NumberT]:
        raise NotImplementedError("Struct members cannot be used as input streams")


@dataclass
class _ScalarDeclarationParams(_DeclarationParams):
    value: Optional[Union[bool, int, float]]

    def create_variable(self, var_name: str, t: type[NumberT]) -> QuaVariable[NumberT]:
        return QuaVariable(var_name, t, init_value=self.value)

    def create_input_stream(self, var_name: str, t: type[NumberT]) -> QuaVariableInputStream[NumberT]:
        return QuaVariableInputStream(var_name, t, init_value=self.value)


def _standardize_value_and_size(
    value: Optional[OneOrMore[Union[bool, int, float]]] = None, size: Optional[int] = None
) -> _DeclarationParams:

    if size is not None:
        size = size.item() if isinstance(size, np.integer) else size
        if not (isinstance(size, int) and size > 0):
            raise QmQuaException("size must be a positive integer")
        if value is not None:
            raise QmQuaException("size declaration cannot be made if value is declared")
        return _ArrayDeclarationParams(size=size, values=[])

    if isinstance(value, Collection):
        size = len(value)
        return _ArrayDeclarationParams(size=size, values=list(value))

    return _ScalarDeclarationParams(value=value)


@overload
def declare(t: type[NumberT]) -> QuaVariable[NumberT]:
    ...


@overload
def declare(t: type[NumberT], value: Literal[None], size: int) -> QuaArrayVariable[NumberT]:
    ...


@overload
def declare(t: type[NumberT], *, size: int) -> QuaArrayVariable[NumberT]:
    ...


@overload
def declare(t: type[NumberT], value: Union[int, bool, float]) -> QuaVariable[NumberT]:
    ...


@overload
def declare(t: type[NumberT], value: Sequence[Union[int, bool, float]]) -> QuaArrayVariable[NumberT]:
    ...


def declare(
    t: type[NumberT],
    value: Optional[OneOrMore[Union[int, bool, float]]] = None,
    size: Optional[int] = None,
) -> Union[QuaVariable[NumberT], QuaArrayVariable[NumberT]]:
    r"""Declare a single QUA variable or QUA vector to be used in subsequent expressions and assignments.

    Declaration is performed by declaring a python variable with the return value of this function.

    Args:
        t: The type of QUA variable. Possible values: ``int``,
            ``fixed``, ``bool``, where:

            ``int``
                a signed 32-bit number
            ``fixed``
                a signed 4.28 fixed point number
            ``bool``
                either `True` or `False`
        value: An initial value for the variable or a list of initial
            values for a vector
        size: If declaring a vector without explicitly specifying a
            value, this parameter is used to specify the length of the
            array

    Returns:
        The variable or vector

    Warning:

        some QUA statements accept a variable with a valid range smaller than the full size of the generic
        QUA variable. For example, ``amp()`` accepts numbers between -2 and 2.
        In case the value stored in the variable is larger than the valid input range, unexpected results
        may occur.

    Example:
        ```python
        a = declare(fixed, value=0.3)
        play('pulse' * amp(a), 'element')

        array1 = declare(int, value=[1, 2, 3])
        array2 = declare(fixed, size=5)
        ```
    """
    params = _standardize_value_and_size(value, size)

    scope = scopes_manager.program_scope
    # We could move the following logic inside the classes, but then we would have to deal with the scope there.
    #  Additionally, we want to separate the concern of choosing the variable name from the declaration logic,
    if isinstance(params, _ArrayDeclarationParams):
        scope.array_index += 1
        var = f"a{scope.array_index}"
    else:
        scope.var_index += 1
        var = f"v{scope.var_index}"

    result = params.create_variable(var, t)
    scope.add_var_declaration(result.declaration_statement)

    return result


def _declare_struct_array_variable(
    t: type[NumberT],
    size: int,
    position: int,
    struct_ref: QuaStructReference,
) -> QuaStructArrayVariable[NumberT, NSize]:
    """
    Declare a struct member variable. This function is used internally when a struct is initialized.
    """
    params = _StructArrayDeclarationParams(size=size, position=position, struct_ref=struct_ref)

    scope = scopes_manager.program_scope
    scope.array_index += 1
    var = f"a{scope.array_index}"

    result: QuaStructArrayVariable[NumberT, NSize] = params.create_variable(var, t)
    scope.add_var_declaration(result.declaration_statement)

    return result


def declare_struct(struct_t: type[StructT]) -> StructT:
    """
    Declare a QUA struct instance inside the current QUA program.

    ``struct_t`` must be a type decorated with [qm.qua.qua_struct][].
    The returned object contains QUA array members that can be assigned, sent to an OPNIC stream with
    [qm.qua.send_to_stream][], or used as the receive target for [qm.qua.receive_from_stream][].

    Example:
        ```python
        @qua_struct
        class Packet:
            data: QuaArray[int, 1]

        with program():
            packet = declare_struct(Packet)
        ```

    Args:
        struct_t (type[StructT]): The QUA struct type to instantiate.

    Returns:
        StructT: A QUA struct instance of the requested type.
    """
    scope = scopes_manager.program_scope
    scope.struct_index += 1
    name = f"s{scope.struct_index}"

    struct_reference = QuaStructReference(name)

    return struct_t(  # type: ignore[call-arg]
        _struct_reference=struct_reference,
        **{
            member_name: factory.create(struct_reference)
            for member_name, factory in struct_t.__members_initializers__.items()
        },
    )


def assign(
    var: Union[QuaArrayCell[NumberT], QuaVariable[NumberT], QuaIO],
    _exp: Union[Scalar[NumberT], QuaIO],
) -> None:
    """Set the value of a given QUA variable, a QUA array cell or an IO to the value of a given expression.

    Args:
        var (QUA variable): A QUA variable, a QUA array cell or an IO for which to assign.
        _exp (QUA expression): An expression for which to set the variable

    Example:
        ```python
        with program() as prog:
            v1 = declare(fixed)
            assign(v1, 1.3)
            play('pulse1' * amp(v1), 'element1')
        ```
    """
    statement = inc_qua_pb2.QuaProgram.AssignmentStatement(
        loc=_get_loc(), target=var.assignment_statement, expression=to_scalar_pb_expression(_exp)
    )
    scopes_manager.append_statement(inc_qua_pb2.QuaProgram.AnyStatement(assign=statement))
    if isinstance(var, DirectStreamSourceInterface):
        var.save()


def _get_save_statement(
    var: ScalarOfAnyType, stream_or_tag: Union[str, ResultStreamSource]
) -> inc_qua_pb2.QuaProgram.AnyStatement:
    if isinstance(stream_or_tag, str):
        result_obj = _declare_save(stream_or_tag, add_legacy_timestamp=True)
    else:
        result_obj = stream_or_tag

    if result_obj.is_adc_trace is True:  # Both None (unset) and False are fine
        raise SaveToAdcTraceException("adc_trace can't be used in save")
    # On a save operation, we set `is_adc_trace` to False if it was previously None.
    # This effectively marks the stream as a non-ADC stream, which affects how it is handled by other commands
    # such as `measure`.
    result_obj.is_adc_trace = False

    statement = inc_qua_pb2.QuaProgram.SaveStatement(
        loc=_get_loc(), source=create_qua_scalar_expression(var).save_statement, tag=result_obj.get_var_name()
    )
    return inc_qua_pb2.QuaProgram.AnyStatement(save=statement)


def save(var: ScalarOfAnyType, stream_or_tag: Union[str, ResultStreamSource]) -> None:
    """Stream a QUA variable, a QUA array cell, or a constant scalar.
    the variable is streamed and not immediately saved (see [Stream processing](../../Guides/stream_proc.md#stream-processing)).
    In case ``stream_or_tag`` is a string, the data will be immediately saved to a result handle under the same name.

    If result variable is used, it can be used in results analysis scope see [stream_processing][qm.qua.stream_processing]
    if string tag is used, it will let you receive result with [qm.QmJob.result_handles][qm.jobs.running_qm_job.RunningQmJob.result_handles].
    Saving to a declared client output stream is equivalent to [qm.qua.send_to_stream][] for scalar values:
    ``save(value, stream)`` and ``send_to_stream(stream, value)`` produce the same client-stream items.
    The type of the variable determines the stream datatype, according to the following rule:

    - int -> int64
    - fixed -> float64
    - bool -> bool

    Note:

        Saving arrays as arrays is not currently supported. Please use a QUA for loop to save an array.

    Example:
        ```python
        # basic save
        a = declare(int, value=2)
        save(a, "a")

        # fetching the results from python (job is a QmJob object):
        a_handle = job.result_handles.get("a")
        a_data = a_handle.fetch_all()

        # save the third array cell
        vec = declare(fixed, value=[0.2, 0.3, 0.4, 0.5])
        save(vec[2], "ArrayCellSave")

        # array iteration
        i = declare(int)
        array = declare(fixed, value=[x / 10 for x in range(30)])
        with for_(i, 0, i < 30, i + 1):
            save(array[i], "array")

        # save a constant
        save(3, "a")
        ```

    Args:
        var (Union[QUA variable, a QUA array cell]): A QUA variable or a
            QUA array cell to save
        stream_or_tag (Union[str, stream variable]): A stream variable
            or string tag name to save the value under
    """
    scopes_manager.append_statement(_get_save_statement(var, stream_or_tag))


def advance_input_stream(
    input_stream: Union[
        QuaVariableInputStream[bool],
        QuaVariableInputStream[int],
        QuaVariableInputStream[float],
        QuaArrayInputStream[bool],
        QuaArrayInputStream[int],
        QuaArrayInputStream[float],
    ],
) -> None:
    """Advances the input stream pointer to the next available variable/vector.

    If there is no new data waiting in the stream, this command will wait until it is available.

    The variable/vector can then be used as a normal QUA variable.

    See [Input streams](../../Guides/features.md#input-streams) for more information.

    -- Available from QOP 2.0 --
    """
    scopes_manager.append_statement(input_stream.advance())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: qm/grpc/v2/common_types.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'qm/grpc/v2/common_types.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dqm/grpc/v2/common_types.proto\x12\nqm.grpc.v2\x1a\x1egoogle/protobuf/wrappers.proto\"u\n\x05Range\x12.\n\x04\x66rom\x18\x01 \x01(\x0b\x32\x1b.google.protobuf.Int64ValueH\x00\x88\x01\x01\x12,\n\x02to\x18\x02 \x01(\x0b\x32\x1b.google.protobuf.Int64ValueH\x01\x88\x01\x01\x42\x07\n\x05_fromB\x05\n\x03_to*\x8b\x01\n\x12JobExecutionStatus\x12\t\n\x05UNSET\x10\x00\x12\x0b\n\x07UNKNOWN\x10\x01\x12\x0b\n\x07PENDING\x10\x02\x12\x0b\n\x07RUNNING\x10\x03\x12\r\n\tCOMPLETED\x10\x04\x12\x0c\n\x08\x43\x41NCELED\x10\x05\x12\x0b\n\x07LOADING\x10\x06\x12\t\n\x05\x45RROR\x10\x07\x12\x0e\n\nPROCESSING\x10\x08\x42\x1a\n\x10qm.client.api.v2P\x01\xa2\x02\x03HLWb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qm.grpc.v2.common_types_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020qm.client.api.v2P\001\242\002\003HLW'
  _globals['_JOBEXECUTIONSTATUS']._serialized_start=197
  _globals['_JOBEXECUTIONSTATUS']._serialized_end=336
  _globals['_RANGE']._serialized_start=77
  _globals['_RANGE']._serialized_end=194
# @@protoc_insertion_point(module_scope)



import abc
import warnings
import dataclasses
from enum import Enum
from dataclasses import dataclass
from typing import Optional, Sequence
from typing import Iterable as IterableClass
from typing import List, Tuple, Union, Literal

import numpy as np
from google.protobuf.struct_pb2 import Value, ListValue

from qm._loc import _get_loc
from qm.type_hinting import Number
from qm.grpc.qm.pb import inc_qua_pb2
from qm.exceptions import QmQuaException
from qm.utils import deprecation_message
from qm.qua._dsl._type_hints import OneOrMore
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._dsl.stream_processing.stream_processing_utils import _ARRAY_SYMBOL, create_array
from qm.qua._expressions import ScalarOfAnyType, OutputStreamInterface, create_qua_scalar_expression
from qm.qua._dsl.stream_processing.map_functions.map_function_classes import (
    FFT,
    DotProduct,
    Convolution,
    BooleanToInt,
    FunctionBase,
    TupleMultiply,
    ConvolutionMode,
    TupleDotProduct,
    MultiplyByScalar,
    MultiplyByVector,
    TupleConvolution,
)


class ResultStream(metaclass=abc.ABCMeta):
    def average(self) -> "UnaryMathOperation":
        """
        Perform a running average on a stream item. The Output of this operation is the running average
        of the values in the stream starting from the beginning of the QUA program.
        """
        return UnaryMathOperation(self, "average")

    def real(self) -> "UnaryMathOperation":
        return UnaryMathOperation(self, "real")

    def image(self) -> "UnaryMathOperation":
        return UnaryMathOperation(self, "image")

    def buffer(self, *args: int) -> "BufferOfStream":
        """Gather items into vectors - creates an array of input stream items and outputs the array as one item.
        only outputs full buffers.

        Note:
            The order of resulting dimensions is different when using a buffer with multiple inputs compared to using
            multiple buffers. The following two lines are equivalent:
            ```python
            stream.buffer(n, l, k)
            stream.buffer(k).buffer(l).buffer(n)
            ```

        Args:
            *args: number of items to gather, can either be a single
                number, which gives the results as a 1d array or
                multiple numbers for a multidimensional array.
        """
        return BufferOfStream(self, *args)

    def buffer_and_skip(self, length: Number, skip: Number) -> "SkippedBufferOfStream":
        """Gather items into vectors - creates an array of input stream items and outputs
        the array as one item.
        Skips the number of given elements. Note that length and skip start from the
        same index, so the `buffer(n)` command is equivalent to `buffer_and_skip(n, n)`.

        Only outputs full buffers.

        Example:
            ```python
            # The stream input is [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]
            with stream_processing():
                stream.buffer(3).save_all("example1")
                stream.buffer_and_skip(3, 3).save_all("example2")
                stream.buffer_and_skip(3, 2).save_all("example3")
                stream.buffer_and_skip(3, 5).save_all("example4")
            # example1 -> [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
            # example2 -> [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
            # example3 -> [[1, 2, 3], [3, 4, 5], [5, 6, 7], [7, 8, 9]]
            # example4 -> [[1, 2, 3], [6, 7, 8]]
            ```

        Args:
            length: number of items to gather
            skip: number of items to skip for each buffer, starting from
                the same index as length
        """
        return SkippedBufferOfStream(self, int(length), int(skip))

    def map(self, function: FunctionBase) -> "MapOfStream":
        """Transform the item by applying a
        [function][qm.qua._dsl.stream_processing.map_functions.map_functions.MapFunctions] to it

        Args:
            function: a function to transform each item to a different
                item. For example, to compute an average between
                elements in a buffer you should write
                ".buffer(len).map(FUNCTIONS.average())"
        """
        return MapOfStream(self, function)

    def flatten(self) -> "UnaryMathOperation":
        """
        Deconstruct an array item - and send its elements one by one as items
        """
        return UnaryMathOperation(self, "flatten")

    def skip(self, length: Number) -> "DiscardedStream":
        """Suppress the first n items of the stream

        Args:
            length: number of items to skip
        """
        return DiscardedStream(self, int(length), "skip")

    def skip_last(self, length: Number) -> "DiscardedStream":
        """Suppress the last n items of the stream

        Args:
            length: number of items to skip
        """
        return DiscardedStream(self, int(length), "skipLast")

    def take(self, length: Number) -> "DiscardedStream":
        """Outputs only the first n items of the stream

        Args:
            length: number of items to take
        """
        return DiscardedStream(self, int(length), "take")

    def histogram(self, bins: List[List[Number]]) -> "HistogramStream":
        """Compute the histogram of all items in stream

        Args:
            bins: vector or pairs. each pair indicates the edge of each
                bin. example: [[1,10],[11,20]] - two bins, one between 1
                and 10, second between 11 and 20
        """
        standardized_bins = [(_bin[0], _bin[1]) for _bin in bins]
        return HistogramStream(self, standardized_bins)

    def zip(self, other: "ResultStream") -> "BinaryOperation":
        """Combine the emissions of two streams to one item that is a tuple of items of input streams

        Args:
            other: second stream to combine with self
        """
        return BinaryOperation(other, self, "zip")

    def save_all(self, tag: str) -> None:
        """Save all items received in stream.

        Args:
            tag: result name
        """
        scopes_manager.append_output_stream(_SaveAllOutputStream(self, tag))

    def save(self, tag: str) -> None:
        """Save only the last item received in stream

        Args:
            tag: result name
        """
        scopes_manager.append_output_stream(_SaveOutputStream(self, tag))

    def _auto_save_all(self, tag: str) -> None:
        """Save all items received in stream.

        Args:
            tag: result name
        """
        scopes_manager.append_output_stream(_AutoSaveAllOutputStream(self, tag))

    def dot_product(self, vector: Sequence[Number]) -> "MapOfStream":
        """Computes dot product of the given vector and each item of the input stream

        Args:
            vector: constant vector of numbers
        """
        return self.map(DotProduct(vector))

    def tuple_dot_product(self) -> "MapOfStream":
        """
        Computes dot product of the given item of the input stream - that should include two vectors
        """
        return self.map(TupleDotProduct())

    def multiply_by(self, scalar_or_vector: OneOrMore[Number]) -> "MapOfStream":
        """Multiply the input stream item by a constant scalar or vector.
        The input item can be either scalar or vector.

        Args:
            scalar_or_vector: either a scalar number, or a vector of
                scalars.
        """
        if isinstance(scalar_or_vector, IterableClass):
            return self.map(MultiplyByVector(scalar_or_vector))
        else:
            return self.map(MultiplyByScalar(scalar_or_vector))

    def tuple_multiply(self) -> "MapOfStream":
        """
        Computes multiplication of the given item of the input stream - that can be any
        combination of scalar and vectors.
        """
        return self.map(TupleMultiply())

    def convolution(self, constant_vector: Sequence[Number], mode: Optional[ConvolutionMode] = "") -> "MapOfStream":
        """Computes discrete, linear convolution of one-dimensional constant vector and one-dimensional vector
        item of the input stream.

        Args:
            constant_vector: vector of numbers
            mode: "full", "same" or "valid"
        """
        if mode is None:
            warnings.warn(
                "mode=None is deprecated, use empty-string or (recommended) don't write the mode at-all.",
                DeprecationWarning,
            )
            mode = ""
        return self.map(Convolution(constant_vector, mode))

    def tuple_convolution(self, mode: Optional[ConvolutionMode] = "") -> "MapOfStream":
        """Computes discrete, linear convolution of two one-dimensional vectors that received as the one item from the input stream

        Args:
            mode: "full", "same" or "valid"
        """
        if mode is None:
            warnings.warn(
                "mode=None is deprecated, use empty-string or (recommended) don't write the mode at-all.",
                DeprecationWarning,
            )
            mode = ""
        return self.map(TupleConvolution(mode))

    def fft(self, output: Optional[str] = None) -> "MapOfStream":
        """Computes one-dimensional discrete fourier transform for every item in the
        stream.
        Item can be a vector of numbers, in this case fft will assume all imaginary
        numbers are 0.
        Item can also be a vector of number pairs - in this case for each pair - the
        first will be real and second imaginary.

        Args:
            output: supported from QOP 1.30 and QOP 2.0, options are
                "normal", "abs" and "angle":

                *   "normal" - Same as default (none), returns a 2d array of
                    size Nx2, where N is the length of the original vector.
                    The first item in each pair is the real part, and the 2nd
                    is the imaginary part.
                *   "abs" - Returns a 1d array of size N with the abs of the fft.
                *   "angle" - Returns the angle between the imaginary and real
                    parts in radians.

        Returns:
            stream object
        """
        return self.map(FFT(output))

    def boolean_to_int(self) -> "MapOfStream":
        """
        converts boolean to an integer number - 1 for true and 0 for false
        """
        return self.map(BooleanToInt())

    def to_proto(self) -> Value:
        return Value(list_value=ListValue(values=list(self._to_list_of_values())))

    @abc.abstractmethod
    def _to_list_of_values(self) -> Sequence[Value]:
        pass

    def add(self, other: Union["ResultStream", OneOrMore[Number]]) -> "BinaryOperation":
        """Allows addition between streams. The addition is done element-wise.
        Can also be performed on buffers and other operators, but they must have the
        same dimensions.

        Example:
            ```python
            i = declare(int)
            j = declare(int)
            k = declare(int, value=5)
            stream = declare_output_stream()
            stream2 = declare_output_stream()
            stream3 = declare_output_stream()
            with for_(j, 0, j < 30, j + 1):
                with for_(i, 0, i < 10, i + 1):
                    save(i, stream)
                    save(j, stream2)
                    save(k, stream3)

            with stream_processing():
                (stream1 + stream2 + stream3).save_all("example1")
                (stream1.buffer(10) + stream2.buffer(10) + stream3.buffer(10)).save_all("example2")
                (stream1 + stream2 + stream3).buffer(10).average().save("example3")
            ```
        """
        return self.__add__(other)

    def subtract(self, other: Union["ResultStream", OneOrMore[Number]]) -> "BinaryOperation":
        """Allows subtraction between streams. The subtraction is done element-wise.
        Can also be performed on buffers and other operators, but they must have the
        same dimensions.

        Example:
            ```python
            i = declare(int)
            j = declare(int)
            k = declare(int, value=5)
            stream = declare_output_stream()
            stream2 = declare_output_stream()
            stream3 = declare_output_stream()
            with for_(j, 0, j < 30, j + 1):
                with for_(i, 0, i < 10, i + 1):
                    save(i, stream)
                    save(j, stream2)
                    save(k, stream3)

            with stream_processing():
                (stream1 - stream2 - stream3).save_all("example1")
                (stream1.buffer(10) - stream2.buffer(10) - stream3.buffer(10)).save_all("example2")
                (stream1 - stream2 - stream3).buffer(10).average().save("example3")
            ```
        """
        return self.__sub__(other)

    def multiply(self, other: Union["ResultStream", OneOrMore[Number]]) -> "BinaryOperation":
        """Allows multiplication between streams. The multiplication is done element-wise.
        Can also be performed on buffers and other operators, but they must have the
        same dimensions.

        Example:
            ```python
            i = declare(int)
            j = declare(int)
            k = declare(int, value=5)
            stream = declare_output_stream()
            stream2 = declare_output_stream()
            stream3 = declare_output_stream()
            with for_(j, 0, j < 30, j + 1):
                with for_(i, 0, i < 10, i + 1):
                    save(i, stream)
                    save(j, stream2)
                    save(k, stream3)

            with stream_processing():
                (stream1 * stream2 * stream3).save_all("example1")
                (stream1.buffer(10) * stream2.buffer(10) * stream3.buffer(10)).save_all("example2")
                (stream1 * stream2 * stream3).buffer(10).average().save("example3")
            ```
        """
        return self.__mul__(other)

    def divide(self, other: Union["ResultStream", OneOrMore[Number]]) -> "BinaryOperation":
        """Allows division between streams. The division is done element-wise.
        Can also be performed on buffers and other operators, but they must have the
        same dimensions.

        Example:
            ```python
            i = declare(int)
            j = declare(int)
            k = declare(int, value=5)
            stream = declare_output_stream()
            stream2 = declare_output_stream()
            stream3 = declare_output_stream()
            with for_(j, 0, j < 30, j + 1):
                with for_(i, 0, i < 10, i + 1):
                    save(i, stream)
                    save(j, stream2)
                    save(k, stream3)

            with stream_processing():
                (stream1 / stream2 / stream3).save_all("example1")
                (stream1.buffer(10) / stream2.buffer(10) / stream3.buffer(10)).save_all("example2")
                (stream1 / stream2 / stream3).buffer(10).average().save("example3")
            ```
        """
        return self.__truediv__(other)

    def __add__(self, other: Union["ResultStream", OneOrMore[Number]]) -> "BinaryOperation":
        return BinaryOperation(self, other, "+")

    def __radd__(self, other: OneOrMore[Number]) -> "BinaryOperation":
        return BinaryOperation(other, self, "+")

    def __sub__(self, other: Union["ResultStream", OneOrMore[Number]]) -> "BinaryOperation":
        return BinaryOperation(self, other, "-")

    def __rsub__(self, other: Union["ResultStream", OneOrMore[Number]]) -> "BinaryOperation":
        return BinaryOperation(other, self, "-")

    def __gt__(self, other: object) -> bool:
        raise QmQuaException("Can't use > operator on results")

    def __ge__(self, other: object) -> bool:
        raise QmQuaException("Can't use >= operator on results")

    def __lt__(self, other: object) -> bool:
        raise QmQuaException("Can't use < operator on results")

    def __le__(self, other: object) -> bool:
        raise QmQuaException("Can't use <= operator on results")

    def __eq__(self, other: object) -> bool:
        raise QmQuaException("Can't use == operator on results")

    def __mul__(self, other: Union["ResultStream", OneOrMore[Number]]) -> "BinaryOperation":
        return BinaryOperation(self, other, "*")

    def __rmul__(self, other: Union["ResultStream", OneOrMore[Number]]) -> "BinaryOperation":
        return BinaryOperation(other, self, "*")

    def __div__(self, other: Union["ResultStream", OneOrMore[Number]]) -> None:
        raise QmQuaException("Can't use / operator on results")

    def __truediv__(self, other: Union["ResultStream", OneOrMore[Number]]) -> "BinaryOperation":
        return BinaryOperation(self, other, "/")

    def __rtruediv__(self, other: Union["ResultStream", OneOrMore[Number]]) -> "BinaryOperation":
        return BinaryOperation(other, self, "/")

    def __lshift__(self, other: ScalarOfAnyType) -> None:
        raise TypeError("Can't use << operator on results of type 'ResultStream', only 'ResultSource'")

    def __rshift__(self, other: Union["ResultStream", OneOrMore[Number]]) -> None:
        raise QmQuaException("Can't use >> operator on results")

    def __and__(self, other: Union["ResultStream", OneOrMore[Number]]) -> None:
        raise QmQuaException("Can't use & operator on results")

    def __or__(self, other: Union["ResultStream", OneOrMore[Number]]) -> None:
        raise QmQuaException("Can't use | operator on results")

    def __xor__(self, other: Union["ResultStream", OneOrMore[Number]]) -> None:
        raise QmQuaException("Can't use ^ operator on results")


class _TimestampMode(Enum):
    Values = 0
    Timestamps = 1
    ValuesAndTimestamps = 2


@dataclass
class _Configuration:
    var_name: str
    timestamp_mode: _TimestampMode
    is_adc_trace: Optional[bool]
    input: int
    auto_reshape: bool
    dtype: Optional[type]


_RESULT_SYMBOL = "@re"


class ResultStreamSource(ResultStream, OutputStreamInterface):
    """A python object representing a source of values that can be processed in a [`stream_processing()`][qm.qua.stream_processing] pipeline

    This interface is chainable, which means that calling most methods on this object will create a new streaming source

    See the base class [ResultStream][qm.qua._dsl.stream_processing.stream_processing.ResultStream] for operations
    """

    def __init__(self, configuration: _Configuration):
        self._configuration = configuration

    @property
    def is_adc_trace(self) -> Optional[bool]:
        return self._configuration.is_adc_trace

    @is_adc_trace.setter
    def is_adc_trace(self, value: bool) -> None:
        self._configuration.is_adc_trace = value

    @property
    def dtype(self) -> Optional[type]:
        return self._configuration.dtype

    def _to_list_of_values(self) -> Sequence[Value]:
        result = [
            Value(string_value=_RESULT_SYMBOL),
            Value(string_value=str(self._configuration.timestamp_mode.value)),
            Value(string_value=self._configuration.var_name),
        ]
        tmp = [
            Value(string_value="@macro_input"),
            Value(string_value=str(self._configuration.input)),
            Value(list_value=ListValue(values=result)),
        ]
        inputs = result if self._configuration.input == -1 else tmp
        macro_auto_reshape = [Value(string_value="@macro_auto_reshape"), Value(list_value=ListValue(values=inputs))]
        auto_reshape = macro_auto_reshape if self._configuration.auto_reshape else inputs
        macro_adc_trace = [Value(string_value="@macro_adc_trace"), Value(list_value=ListValue(values=auto_reshape))]
        return macro_adc_trace if self.is_adc_trace else auto_reshape

    def get_var_name(self) -> str:
        return self._configuration.var_name

    def with_timestamps(self) -> "ResultStreamSource":
        """Get a stream with the relevant timestamp for each stream-item"""
        return ResultStreamSource(
            dataclasses.replace(
                self._configuration,
                timestamp_mode=_TimestampMode.ValuesAndTimestamps,
            )
        )

    def timestamps(self) -> "ResultStreamSource":
        """Get a stream with only the timestamps of the stream-items"""
        warnings.warn(
            deprecation_message(
                method="timestamps",
                deprecated_in="1.2.3",
                removed_in="2.0.0",
                details="timestamps() is deprecated, please use with_timestamps() instead.",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return ResultStreamSource(
            dataclasses.replace(
                self._configuration,
                timestamp_mode=_TimestampMode.Timestamps,
            )
        )

    def input1(self) -> "ResultStreamSource":
        """A stream of raw ADC data from input 1. Only relevant when saving data from measure statement."""
        return ResultStreamSource(dataclasses.replace(self._configuration, input=1))

    def input2(self) -> "ResultStreamSource":
        """A stream of raw ADC data from input 2. Only relevant when saving data from measure statement."""
        return ResultStreamSource(dataclasses.replace(self._configuration, input=2))

    def auto_reshape(self) -> "ResultStreamSource":
        """Creates a buffer with dimensions according to the program structure in QUA.

        For example, when running the following program the result "reshaped" will have
        shape of (30,10):

        Example:
            ```python
            i = declare(int)
            j = declare(int)
            stream = declare_output_stream()
            with for_(i, 0, i < 30, i + 1):
                with for_(j, 0, j < 10, j + 1):
                    save(i, stream)

            with stream_processing():
                stream.auto_reshape().save_all("reshaped")
            ```
        """
        return ResultStreamSource(dataclasses.replace(self._configuration, auto_reshape=True))

    def __lshift__(self, other: ScalarOfAnyType) -> None:
        if self.is_adc_trace is True:  # Both None (unset) and False are fine
            raise QmQuaException("adc_trace can't be used in save")
        # Since a lshift is equivalent to a save operation, we set `is_adc_trace` to False if it was previously None.
        # This effectively marks the stream as a non-ADC stream, which affects how it is handled by other commands
        # such as `measure`.
        self.is_adc_trace = False

        statement = inc_qua_pb2.QuaProgram.SaveStatement(
            loc=_get_loc(), source=create_qua_scalar_expression(other).save_statement, tag=self.get_var_name()
        )
        scopes_manager.append_statement(inc_qua_pb2.QuaProgram.AnyStatement(save=statement))


class _UnaryOperation(ResultStream, metaclass=abc.ABCMeta):
    def __init__(self, input_stream: "ResultStream"):
        self._input_stream = input_stream

    def _to_list_of_values(self) -> Sequence[Value]:
        return [Value(string_value=self._operator_name)] + list(self._args) + [self._input_stream.to_proto()]

    @property
    @abc.abstractmethod
    def _operator_name(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def _args(self) -> Sequence[Value]:
        pass


class UnaryMathOperation(_UnaryOperation):
    def __init__(
        self, input_stream: "ResultStream", operator_name: Literal["average", "real", "image", "flatten", "tmult"]
    ) -> None:
        super().__init__(input_stream)
        self._operator = operator_name

    @property
    def _operator_name(self) -> str:
        return self._operator

    @property
    def _args(self) -> Sequence[Value]:
        return []


class BufferOfStream(_UnaryOperation):
    def __init__(self, input_stream: "ResultStream", *args: int):
        super().__init__(input_stream)
        self._args_input = args

    @property
    def _args(self) -> List[Value]:
        return [Value(string_value=str(int(arg))) for arg in self._args_input]

    @property
    def _operator_name(self) -> str:
        return "buffer"


class SkippedBufferOfStream(_UnaryOperation):
    def __init__(self, input_stream: "ResultStream", length: int, skip: int):
        super().__init__(input_stream)
        self._length = length
        self._skip = skip

    @property
    def _args(self) -> List[Value]:
        return [Value(string_value=str(self._length)), Value(string_value=str(self._skip))]

    @property
    def _operator_name(self) -> str:
        return "bufferAndSkip"


class MapOfStream(_UnaryOperation):
    def __init__(self, input_stream: "ResultStream", function: FunctionBase):
        super().__init__(input_stream)
        self._function = function

    @property
    def _operator_name(self) -> str:
        return "map"

    @property
    def _args(self) -> List[Value]:
        return [self._function.to_proto()]


class DiscardedStream(_UnaryOperation):
    def __init__(self, input_stream: "ResultStream", length: int, operator_name: Literal["skip", "skipLast", "take"]):
        super().__init__(input_stream)
        self._length = length
        self._operator_input = operator_name

    @property
    def _operator_name(self) -> str:
        return self._operator_input

    @property
    def _args(self) -> List[Value]:
        return [Value(string_value=str(self._length))]


class HistogramStream(_UnaryOperation):
    def __init__(self, input_stream: "ResultStream", bins_: Sequence[Tuple[Number, Number]]):
        super().__init__(input_stream)
        self._bins = bins_

    @property
    def _operator_name(self) -> str:
        return "histogram"

    @property
    def _args(self) -> List[Value]:
        tmp = [Value(string_value=_ARRAY_SYMBOL)]
        converted_bins = [create_array(sub_list) for sub_list in self._bins]
        _bins = Value(list_value=ListValue(values=tmp + converted_bins))
        return [_bins]


class BinaryOperation(ResultStream):
    def __init__(
        self,
        lhs: Union["ResultStream", OneOrMore[Number]],
        rhs: Union["ResultStream", OneOrMore[Number]],
        operator_name: Literal["+", "-", "*", "/", "zip"],
    ):
        self._lhs = lhs
        self._rhs = rhs
        self._operator_name = operator_name

    def _standardize_output(self, other: Union["ResultStream", OneOrMore[Number]]) -> Value:
        if isinstance(other, ResultStream):
            return other.to_proto()
        elif isinstance(other, (int, float, np.integer, np.floating)) and not isinstance(other, (bool, np.bool_)):
            return Value(string_value=str(other))
        elif isinstance(other, IterableClass):
            return create_array(other)
        if self._operator_name == "zip":
            raise TypeError(f"Unsupported zip for '{type(self._lhs)} and {type(self._rhs)}.")
        else:
            raise TypeError(f"Unsupported operation - '{type(self._lhs)} {self._operator_name} {type(self._rhs)}.")

    def _to_list_of_values(self) -> Sequence[Value]:
        return [
            Value(string_value=self._operator_name),
            self._standardize_output(self._lhs),
            self._standardize_output(self._rhs),
        ]


class _OutputStream(metaclass=abc.ABCMeta):
    """Even though it looks like a stream, it does not support operations (like __add__) and hence, a different object"""

    def __init__(self, input_stream: "ResultStream", tag: str):
        self._input_stream = input_stream
        self.tag = tag

    def to_proto(self) -> "ListValue":
        values = [Value(string_value=s) for s in self._operator_array]
        values.append(self._input_stream.to_proto())
        return ListValue(values=values)

    @property
    @abc.abstractmethod
    def _operator_array(self) -> Tuple[str, ...]:
        pass


class _SaveOutputStream(_OutputStream):
    @property
    def _operator_array(self) -> Tuple[str, ...]:
        return "save", self.tag


class _SaveAllOutputStream(_OutputStream):
    @property
    def _operator_array(self) -> Tuple[str, ...]:
        return "saveAll", self.tag


class _AutoSaveAllOutputStream(_OutputStream):
    @property
    def _operator_array(self) -> Tuple[str, ...]:
        return "saveAll", self.tag, "auto"


StreamType = Union[str, ResultStreamSource]
"""A type for a stream object in QUA."""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: octave/v1/api.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'octave/v1/api.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13octave/v1/api.proto\x12\x1dquantummachines.octave.api.v1\x1a\x1egoogle/protobuf/wrappers.proto\"\x17\n\x15SubscribeStateRequest\"\x8e\x01\n\x16SubscribeStateResponse\x12\x39\n\x05state\x18\x01 \x01(\x0e\x32*.quantummachines.octave.api.v1.DeviceState\x12\x39\n\x05\x65rror\x18\x02 \x01(\x0e\x32*.quantummachines.octave.api.v1.DeviceError\"\x0e\n\x0cResetRequest\":\n\rResetResponse\x12\x12\n\nis_success\x18\x01 \x01(\x08\x12\x15\n\rerror_message\x18\x02 \x01(\t\"F\n\rHealthRequest\x12 \n\x18monitor_interval_seconds\x18\x01 \x01(\r\x12\x13\n\x0bstop_stream\x18\x02 \x01(\x08\"\x92\x01\n\x0eHealthResponse\x12?\n\x07\x65xplore\x18\x01 \x01(\x0b\x32..quantummachines.octave.api.v1.ExploreResponse\x12?\n\x07monitor\x18\x02 \x01(\x0b\x32..quantummachines.octave.api.v1.MonitorResponse\"\x13\n\x11GetVersionRequest\"%\n\x12GetVersionResponse\x12\x0f\n\x07version\x18\x01 \x01(\t\"[\n\x0fModuleReference\x12\x39\n\x04type\x18\x01 \x01(\x0e\x32+.quantummachines.octave.api.v1.OctaveModule\x12\r\n\x05index\x18\x02 \x01(\x05\"\x80\x01\n\x0bSaveRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12?\n\x07modules\x18\x02 \x03(\x0b\x32..quantummachines.octave.api.v1.ModuleReference\x12\x11\n\toverwrite\x18\x03 \x01(\x08\x12\x11\n\ttimestamp\x18\x04 \x01(\x03\"0\n\x0cSaveResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"\x1b\n\rRecallRequest\x12\n\n\x02id\x18\x01 \x01(\t\"8\n\x0eRecallResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x15\n\rerror_message\x18\x02 \x01(\t\"\r\n\x0bListRequest\"\x86\x01\n\x08SaveInfo\x12;\n\x07request\x18\x01 \x01(\x0b\x32*.quantummachines.octave.api.v1.SaveRequest\x12=\n\x07\x63ontent\x18\x02 \x01(\x0b\x32,.quantummachines.octave.api.v1.UpdateRequest\"K\n\x0cListResponse\x12;\n\nsave_infos\x18\x01 \x03(\x0b\x32\'.quantummachines.octave.api.v1.SaveInfo\"$\n\x0eMonitorRequest\x12\x12\n\nsense_only\x18\x01 \x01(\x08\"\xf8\x05\n\x0fMonitorResponse\x12L\n\x07modules\x18\x01 \x03(\x0b\x32;.quantummachines.octave.api.v1.MonitorResponse.ModuleStatus\x1a]\n\x11ModuleStatusError\x12H\n\x04type\x18\x01 \x01(\x0e\x32:.quantummachines.octave.api.v1.MonitorResponse.OctaveError\x1a\xb5\x01\n\x0cModuleStatus\x12>\n\x06module\x18\x01 \x01(\x0b\x32..quantummachines.octave.api.v1.ModuleReference\x12\x13\n\x0btemperature\x18\x02 \x01(\x02\x12P\n\x06\x65rrors\x18\x03 \x03(\x0b\x32@.quantummachines.octave.api.v1.MonitorResponse.ModuleStatusError\"\xff\x02\n\x0bOctaveError\x12\x1c\n\x18OCTAVE_ERROR_UNSPECIFIED\x10\x00\x12#\n\x1fOCTAVE_ERROR_BAD_EXTERNAL_CLOCK\x10\x01\x12#\n\x1fOCTAVE_ERROR_BAD_INTERNAL_CLOCK\x10\x02\x12\x1f\n\x1bOCTAVE_ERROR_FPGA_NOT_FOUND\x10\x03\x12\x1d\n\x19OCTAVE_ERROR_FPGA_HW_FAIL\x10\x04\x12\"\n\x1eOCTAVE_ERROR_FPGA_BITFILE_FAIL\x10\x05\x12!\n\x1dOCTAVE_ERROR_MODULE_NOT_FOUND\x10\x06\x12\x19\n\x15OCTAVE_ERROR_OVERHEAT\x10\x07\x12$\n OCTAVE_ERROR_OVERHEAT_PROTECTION\x10\x08\x12\x1f\n\x1bOCTAVE_ERROR_READ_TEMP_FAIL\x10\t\x12\x1f\n\x1bOCTAVE_ERROR_PLL_NOT_LOCKED\x10\n\"2\n\x0e\x43ontrolRequest\x12\x0e\n\x06w_data\x18\x01 \x01(\x0c\x12\x10\n\x08r_length\x18\x02 \x01(\x05\"\x8f\x01\n\x0f\x43ontrolResponse\x12\x0e\n\x06r_data\x18\x01 \x01(\x0c\"l\n\nRdataDebug\x12\x1b\n\x17RDATA_DEBUG_UNSPECIFIED\x10\x00\x12\x1f\n\x1aRDATA_DEBUG_ERROR_RESPONSE\x10\xa0\x01\x12 \n\x1cRDATA_DEBUG_SUCCESS_RESPONSE\x10 \"\x10\n\x0e\x45xploreRequest\"\xb3\x01\n\x0f\x45xploreResponse\x12H\n\x07modules\x18\x01 \x03(\x0b\x32\x37.quantummachines.octave.api.v1.ExploreResponse.ModuleId\x1aV\n\x08ModuleId\x12>\n\x06module\x18\x01 \x01(\x0b\x32..quantummachines.octave.api.v1.ModuleReference\x12\n\n\x02id\x18\x02 \x01(\t\"M\n\rUpdateRequest\x12<\n\x07updates\x18\x01 \x03(\x0b\x32+.quantummachines.octave.api.v1.SingleUpdate\"8\n\x0eUpdateResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x15\n\rerror_message\x18\x02 \x01(\t\"\xb2\x03\n\x0cSingleUpdate\x12\x43\n\nrf_up_conv\x18\x01 \x01(\x0b\x32-.quantummachines.octave.api.v1.RFUpConvUpdateH\x00\x12G\n\x0crf_down_conv\x18\x02 \x01(\x0b\x32/.quantummachines.octave.api.v1.RFDownConvUpdateH\x00\x12G\n\x0cif_down_conv\x18\x03 \x01(\x0b\x32/.quantummachines.octave.api.v1.IFDownConvUpdateH\x00\x12;\n\x05synth\x18\x04 \x01(\x0b\x32*.quantummachines.octave.api.v1.SynthUpdateH\x00\x12;\n\x05\x63lock\x18\x05 \x01(\x0b\x32*.quantummachines.octave.api.v1.ClockUpdateH\x00\x12G\n\x0bmotherboard\x18\x06 \x01(\x0b\x32\x30.quantummachines.octave.api.v1.MotherboardUpdateH\x00\x42\x08\n\x06update\"\xc7\x05\n\x0eRFUpConvUpdate\x12\r\n\x05index\x18\x01 \x01(\r\x12R\n\ninput_attn\x18\x02 \x01(\x0e\x32>.quantummachines.octave.api.v1.RFUpConvUpdate.IfInputSelection\x12+\n\x07\x65nabled\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12\x37\n\x11mixer_output_attn\x18\x04 \x01(\x0b\x32\x1c.google.protobuf.UInt32Value\x12\x35\n\x11power_amp_enabled\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12\x34\n\x0epower_amp_attn\x18\x06 \x01(\x0b\x32\x1c.google.protobuf.UInt32Value\x12V\n\x10\x66\x61st_switch_mode\x18\x07 \x01(\x0e\x32<.quantummachines.octave.api.v1.RFUpConvUpdate.FastSwitchMode\"\x82\x01\n\x10IfInputSelection\x12\"\n\x1eIF_INPUT_SELECTION_UNSPECIFIED\x10\x00\x12#\n\x1fIF_INPUT_SELECTION_PASS_THROUGH\x10\x01\x12%\n!IF_INPUT_SELECTION_ATTENUATE_10DB\x10\x02\"\xa1\x01\n\x0e\x46\x61stSwitchMode\x12 \n\x1c\x46\x41ST_SWITCH_MODE_UNSPECIFIED\x10\x00\x12\x17\n\x13\x46\x41ST_SWITCH_MODE_ON\x10\x01\x12\x18\n\x14\x46\x41ST_SWITCH_MODE_OFF\x10\x02\x12\x1b\n\x17\x46\x41ST_SWITCH_MODE_DIRECT\x10\x03\x12\x1d\n\x19\x46\x41ST_SWITCH_MODE_INVERTED\x10\x04\"\x9f\x04\n\x10RFDownConvUpdate\x12\r\n\x05index\x18\x01 \x01(\r\x12I\n\x08rf_input\x18\x02 \x01(\x0e\x32\x37.quantummachines.octave.api.v1.RFDownConvUpdate.RFInput\x12I\n\x08lo_input\x18\x03 \x01(\x0e\x32\x37.quantummachines.octave.api.v1.RFDownConvUpdate.LOInput\x12+\n\x07\x65nabled\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\"\xd8\x01\n\x07RFInput\x12\x18\n\x14RF_INPUT_UNSPECIFIED\x10\x00\x12\x14\n\x10RF_INPUT_DEBUG_1\x10\x01\x12\x14\n\x10RF_INPUT_DEBUG_2\x10\x02\x12\x14\n\x10RF_INPUT_DEBUG_3\x10\x03\x12\x14\n\x10RF_INPUT_DEBUG_4\x10\x04\x12\x14\n\x10RF_INPUT_DEBUG_5\x10\x05\x12\x11\n\rRF_INPUT_MAIN\x10\x06\x12\x19\n\x15RF_INPUT_FPGA_CONTROL\x10\x07\x12\x17\n\x13RF_INPUT_DISCONNECT\x10\x08\"^\n\x07LOInput\x12\x18\n\x14LO_INPUT_UNSPECIFIED\x10\x00\x12\x0e\n\nLO_INPUT_1\x10\x01\x12\x0e\n\nLO_INPUT_2\x10\x02\x12\x19\n\x15LO_INPUT_FPGA_CONTROL\x10\x03\"\x92\x04\n\x10IFDownConvUpdate\x12\r\n\x05index\x18\x01 \x01(\r\x12I\n\x08\x63hannel1\x18\x02 \x01(\x0b\x32\x37.quantummachines.octave.api.v1.IFDownConvUpdate.Channel\x12I\n\x08\x63hannel2\x18\x03 \x01(\x0b\x32\x37.quantummachines.octave.api.v1.IFDownConvUpdate.Channel\x1a\x99\x01\n\x07\x43hannel\x12\x42\n\x04mode\x18\x01 \x01(\x0e\x32\x34.quantummachines.octave.api.v1.IFDownConvUpdate.Mode\x12J\n\x08\x63oupling\x18\x02 \x01(\x0e\x32\x38.quantummachines.octave.api.v1.IFDownConvUpdate.Coupling\"b\n\x04Mode\x12\x14\n\x10MODE_UNSPECIFIED\x10\x00\x12\x0c\n\x08MODE_OFF\x10\x01\x12\x0f\n\x0bMODE_BYPASS\x10\x02\x12\x15\n\x11MODE_POWER_DETECT\x10\x03\x12\x0e\n\nMODE_MIXER\x10\x04\"Y\n\x08\x43oupling\x12\x18\n\x14\x43OUPLING_UNSPECIFIED\x10\x00\x12\x11\n\rCOUPLING_OPEN\x10\x01\x12\x0f\n\x0b\x43OUPLING_DC\x10\x02\x12\x0f\n\x0b\x43OUPLING_AC\x10\x03\"\xba\x02\n\x0b\x43lockUpdate\x12=\n\x04mode\x18\x02 \x01(\x0e\x32/.quantummachines.octave.api.v1.ClockUpdate.Mode\x12\x35\n\x0f\x63lock_frequency\x18\x03 \x01(\x0b\x32\x1c.google.protobuf.DoubleValue\x12\x38\n\x12synthesizers_clock\x18\x04 \x01(\x0b\x32\x1c.google.protobuf.DoubleValue\x12\x16\n\tclustered\x18\x05 \x01(\x08H\x00\x88\x01\x01\"U\n\x04Mode\x12\x14\n\x10MODE_UNSPECIFIED\x10\x00\x12\x11\n\rMODE_BUFFERED\x10\x01\x12\x11\n\rMODE_EXTERNAL\x10\x02\x12\x11\n\rMODE_INTERNAL\x10\x03\x42\x0c\n\n_clustered\"D\n\x11MotherboardUpdate\x12/\n\tfan_speed\x18\x01 \x01(\x0b\x32\x1c.google.protobuf.DoubleValue\"\xaf\x10\n\x0bSynthUpdate\x12\r\n\x05index\x18\x01 \x01(\r\x12R\n\x0freference_clock\x18\x02 \x01(\x0b\x32\x39.quantummachines.octave.api.v1.SynthUpdate.ReferenceClock\x12L\n\x0csynth_output\x18\x03 \x01(\x0b\x32\x36.quantummachines.octave.api.v1.SynthUpdate.SynthOutput\x12W\n\x12synth_output_power\x18\x04 \x01(\x0e\x32;.quantummachines.octave.api.v1.SynthUpdate.SynthOutputPower\x12,\n\x06heater\x18\x05 \x01(\x0b\x32\x1c.google.protobuf.UInt32Value\x12*\n\x04gain\x18\x06 \x01(\x0b\x32\x1c.google.protobuf.UInt32Value\x12\x32\n\x0c\x64igital_attn\x18\x07 \x01(\x0b\x32\x1c.google.protobuf.UInt32Value\x12J\n\x0bmain_source\x18\x08 \x01(\x0e\x32\x35.quantummachines.octave.api.v1.SynthUpdate.MainSource\x12J\n\x0bmain_output\x18\t \x01(\x0e\x32\x35.quantummachines.octave.api.v1.SynthUpdate.MainOutput\x12T\n\x10secondary_output\x18\n \x01(\x0e\x32:.quantummachines.octave.api.v1.SynthUpdate.SecondaryOutput\x12I\n\nstabilizer\x18\x0b \x01(\x0b\x32\x35.quantummachines.octave.api.v1.SynthUpdate.Stabilizer\x1a\xbb\x01\n\x0eReferenceClock\x12J\n\x06source\x18\x01 \x01(\x0e\x32:.quantummachines.octave.api.v1.SynthUpdate.ReferenceSource\x12/\n\tfrequency\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.DoubleValue\x12,\n\x07\x64ivider\x18\x03 \x01(\x0b\x32\x1b.google.protobuf.Int32Value\x1az\n\x0bSynthOutput\x12\x31\n\tfrequency\x18\x01 \x01(\x0b\x32\x1c.google.protobuf.DoubleValueH\x00\x12.\n\x08\x64isabled\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.BoolValueH\x00\x42\x08\n\x06update\x1a\xe9\x02\n\nStabilizer\x12(\n\x03k_p\x18\x01 \x01(\x0b\x32\x1b.google.protobuf.FloatValue\x12(\n\x03k_i\x18\x02 \x01(\x0b\x32\x1b.google.protobuf.FloatValue\x12(\n\x03k_d\x18\x03 \x01(\x0b\x32\x1b.google.protobuf.FloatValue\x12&\n\x01\x61\x18\x04 \x01(\x0b\x32\x1b.google.protobuf.FloatValue\x12&\n\x01\x62\x18\x05 \x01(\x0b\x32\x1b.google.protobuf.FloatValue\x12\x30\n\nheater_max\x18\x06 \x01(\x0b\x32\x1c.google.protobuf.UInt32Value\x12.\n\tset_point\x18\x07 \x01(\x0b\x32\x1b.google.protobuf.FloatValue\x12+\n\x07\x65nabled\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\"q\n\x0fReferenceSource\x12 \n\x1cREFERENCE_SOURCE_UNSPECIFIED\x10\x00\x12\x1d\n\x19REFERENCE_SOURCE_INTERNAL\x10\x01\x12\x1d\n\x19REFERENCE_SOURCE_EXTERNAL\x10\x02\"\xb2\x01\n\x10SynthOutputPower\x12\"\n\x1eSYNTH_OUTPUT_POWER_UNSPECIFIED\x10\x00\x12\x1d\n\x19SYNTH_OUTPUT_POWER_NEG4DB\x10\x01\x12\x1d\n\x19SYNTH_OUTPUT_POWER_NEG1DB\x10\x02\x12\x1d\n\x19SYNTH_OUTPUT_POWER_POS2DB\x10\x03\x12\x1d\n\x19SYNTH_OUTPUT_POWER_POS5DB\x10\x04\"`\n\nMainSource\x12\x1b\n\x17MAIN_SOURCE_UNSPECIFIED\x10\x00\x12\x18\n\x14MAIN_SOURCE_EXTERNAL\x10\x01\x12\x1b\n\x17MAIN_SOURCE_SYNTHESIZER\x10\x02\"r\n\nMainOutput\x12\x1b\n\x17MAIN_OUTPUT_UNSPECIFIED\x10\x00\x12\x14\n\x10MAIN_OUTPUT_MAIN\x10\x01\x12\x13\n\x0fMAIN_OUTPUT_OFF\x10\x02\x12\x1c\n\x18MAIN_OUTPUT_FPGA_CONTROL\x10\x03\"\xaa\x01\n\x0fSecondaryOutput\x12 \n\x1cSECONDARY_OUTPUT_UNSPECIFIED\x10\x00\x12\x19\n\x15SECONDARY_OUTPUT_MAIN\x10\x01\x12\x1d\n\x19SECONDARY_OUTPUT_AUXILARY\x10\x02\x12\x18\n\x14SECONDARY_OUTPUT_OFF\x10\x03\x12!\n\x1dSECONDARY_OUTPUT_FPGA_CONTROL\x10\x04\"c\n\rAquireRequest\x12?\n\x07modules\x18\x01 \x03(\x0b\x32..quantummachines.octave.api.v1.ModuleReference\x12\x11\n\tuse_cache\x18\x02 \x01(\x08\"M\n\x0e\x41quireResponse\x12;\n\x05state\x18\x01 \x01(\x0b\x32,.quantummachines.octave.api.v1.UpdateRequest\"\x11\n\x0fIdentifyRequest\"\x85\x03\n\x10IdentifyResponse\x12I\n\x10rf_up_converters\x18\x01 \x03(\x0b\x32/.quantummachines.octave.api.v1.RFUpConvIdentity\x12M\n\x12rf_down_converters\x18\x02 \x03(\x0b\x32\x31.quantummachines.octave.api.v1.RFDownConvIdentity\x12M\n\x12if_down_converters\x18\x03 \x03(\x0b\x32\x31.quantummachines.octave.api.v1.IFDownConvIdentity\x12\x42\n\x0csynthesizers\x18\x04 \x03(\x0b\x32,.quantummachines.octave.api.v1.SynthIdentity\x12\x44\n\x0epanel_identity\x18\x05 \x01(\x0b\x32,.quantummachines.octave.api.v1.PanelIdentity\"\x1f\n\x0eUpConvRFOutput\x12\r\n\x05index\x18\x01 \x01(\r\"\xdf\x01\n\rSynthRFOutput\x12\r\n\x05index\x18\x01 \x01(\r\x12L\n\x0boutput_port\x18\x02 \x01(\x0e\x32\x37.quantummachines.octave.api.v1.SynthRFOutput.OutputPort\"q\n\nOutputPort\x12\x1b\n\x17OUTPUT_PORT_UNSPECIFIED\x10\x00\x12\x15\n\x11OUTPUT_PORT_SYNTH\x10\x01\x12\x14\n\x10OUTPUT_PORT_MAIN\x10\x02\x12\x19\n\x15OUTPUT_PORT_SECONDARY\x10\x03\"k\n\x0f\x45xternalRFInput\x12\x18\n\x0elo_input_index\x18\x01 \x01(\rH\x00\x12\x1e\n\x14\x64\x65mod_lo_input_index\x18\x02 \x01(\rH\x00\x12\x15\n\x0brf_in_index\x18\x03 \x01(\rH\x00\x42\x07\n\x05input\"\xba\x02\n\x08RFSource\x12J\n\x11rf_up_conv_output\x18\x01 \x01(\x0b\x32-.quantummachines.octave.api.v1.UpConvRFOutputH\x00\x12\x44\n\x0csynth_output\x18\x02 \x01(\x0b\x32,.quantummachines.octave.api.v1.SynthRFOutputH\x00\x12H\n\x0e\x65xternal_input\x18\x03 \x01(\x0b\x32..quantummachines.octave.api.v1.ExternalRFInputH\x00\x12H\n\x0f\x63onstant_source\x18\x04 \x01(\x0e\x32-.quantummachines.octave.api.v1.ConstantSourceH\x00\x42\x08\n\x06source\"\xc7\x01\n\x12RFDownConvIFSource\x12\r\n\x05index\x18\x01 \x01(\r\x12Q\n\x0boutput_port\x18\x02 \x01(\x0e\x32<.quantummachines.octave.api.v1.RFDownConvIFSource.OutputPort\"O\n\nOutputPort\x12\x1b\n\x17OUTPUT_PORT_UNSPECIFIED\x10\x00\x12\x11\n\rOUTPUT_PORT_I\x10\x01\x12\x11\n\rOUTPUT_PORT_Q\x10\x02\":\n\x12IFDownConvIFSource\x12\r\n\x05index\x18\x01 \x01(\r\x12\x15\n\rchannel_index\x18\x02 \x01(\r\"r\n\x0f\x45xternalIFInput\x12\x11\n\x07i_index\x18\x01 \x01(\rH\x00\x12\x11\n\x07q_index\x18\x02 \x01(\rH\x00\x12\x17\n\rif_lo_i_index\x18\x03 \x01(\rH\x00\x12\x17\n\rif_lo_q_index\x18\x04 \x01(\rH\x00\x42\x07\n\x05input\"\xd3\x02\n\x08IFSource\x12R\n\x15rf_downconv_if_source\x18\x01 \x01(\x0b\x32\x31.quantummachines.octave.api.v1.RFDownConvIFSourceH\x00\x12R\n\x15if_downconv_if_source\x18\x02 \x01(\x0b\x32\x31.quantummachines.octave.api.v1.IFDownConvIFSourceH\x00\x12K\n\x11\x65xternal_if_input\x18\x03 \x01(\x0b\x32..quantummachines.octave.api.v1.ExternalIFInputH\x00\x12H\n\x0f\x63onstant_source\x18\x04 \x01(\x0e\x32-.quantummachines.octave.api.v1.ConstantSourceH\x00\x42\x08\n\x06source\"\xb9\x03\n\x10RFUpConvIdentity\x12\r\n\x05index\x18\x01 \x01(\r\x12R\n\x0c\x63onnectivity\x18\x02 \x01(\x0b\x32<.quantummachines.octave.api.v1.RFUpConvIdentity.Connectivity\x12N\n\nparameters\x18\x03 \x01(\x0b\x32:.quantummachines.octave.api.v1.RFUpConvIdentity.Parameters\x1a\xbd\x01\n\x0c\x43onnectivity\x12\x38\n\x07i_input\x18\x01 \x01(\x0b\x32\'.quantummachines.octave.api.v1.IFSource\x12\x38\n\x07q_input\x18\x02 \x01(\x0b\x32\'.quantummachines.octave.api.v1.IFSource\x12\x39\n\x08lo_input\x18\x03 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\x1a\x32\n\nParameters\x12\x11\n\tattn_1_db\x18\x01 \x01(\x02\x12\x11\n\tattn_2_db\x18\x02 \x01(\x02\"\x93\x05\n\x12RFDownConvIdentity\x12\r\n\x05index\x18\x01 \x01(\r\x12T\n\x0c\x63onnectivity\x18\x02 \x01(\x0b\x32>.quantummachines.octave.api.v1.RFDownConvIdentity.Connectivity\x1a\x97\x04\n\x0c\x43onnectivity\x12\x41\n\x10\x64\x65\x62ug_rf_input_1\x18\x01 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\x12\x41\n\x10\x64\x65\x62ug_rf_input_2\x18\x02 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\x12\x41\n\x10\x64\x65\x62ug_rf_input_3\x18\x03 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\x12\x41\n\x10\x64\x65\x62ug_rf_input_4\x18\x04 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\x12\x41\n\x10\x64\x65\x62ug_rf_input_5\x18\x05 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\x12>\n\rrf_main_input\x18\x06 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\x12;\n\nlo_input_1\x18\x07 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\x12;\n\nlo_input_2\x18\x08 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\"\x98\x03\n\x12IFDownConvIdentity\x12\r\n\x05index\x18\x01 \x01(\r\x12T\n\x0c\x63onnectivity\x18\x02 \x01(\x0b\x32>.quantummachines.octave.api.v1.IFDownConvIdentity.Connectivity\x1a\x9c\x02\n\x0c\x43onnectivity\x12@\n\x0f\x63hannel_1_input\x18\x01 \x01(\x0b\x32\'.quantummachines.octave.api.v1.IFSource\x12\x43\n\x12\x63hannel_1_lo_input\x18\x02 \x01(\x0b\x32\'.quantummachines.octave.api.v1.IFSource\x12@\n\x0f\x63hannel_2_input\x18\x03 \x01(\x0b\x32\'.quantummachines.octave.api.v1.IFSource\x12\x43\n\x12\x63hannel_2_lo_input\x18\x04 \x01(\x0b\x32\'.quantummachines.octave.api.v1.IFSource\"\x83\x06\n\rSynthIdentity\x12\r\n\x05index\x18\x01 \x01(\r\x12O\n\x0c\x63onnectivity\x18\x02 \x01(\x0b\x32\x39.quantummachines.octave.api.v1.SynthIdentity.Connectivity\x12K\n\nparameters\x18\x03 \x01(\x0b\x32\x37.quantummachines.octave.api.v1.SynthIdentity.Parameters\x1a\x93\x01\n\x0c\x43onnectivity\x12>\n\rmain_lo_input\x18\x01 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\x12\x43\n\x12secondary_lo_input\x18\x02 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\x1a\xae\x03\n\nParameters\x12i\n\x15low_frequency_filters\x18\x01 \x03(\x0b\x32J.quantummachines.octave.api.v1.SynthIdentity.Parameters.LowFrequencyFilter\x12k\n\x17medium_frequency_filter\x18\x02 \x01(\x0b\x32J.quantummachines.octave.api.v1.SynthIdentity.Parameters.ParametrizedFilter\x12i\n\x15high_frequency_filter\x18\x03 \x01(\x0b\x32J.quantummachines.octave.api.v1.SynthIdentity.Parameters.ParametrizedFilter\x1aG\n\x12LowFrequencyFilter\x12\r\n\x05index\x18\x01 \x01(\r\x12\x10\n\x08\x66ilter_1\x18\x02 \x01(\t\x12\x10\n\x08\x66ilter_2\x18\x03 \x01(\t\x1a\x14\n\x12ParametrizedFilter\"\xd2\x03\n\rPanelIdentity\x12I\n\nrf_outputs\x18\x01 \x03(\x0b\x32\x35.quantummachines.octave.api.v1.PanelIdentity.RFOutput\x12<\n\x0bif_output_i\x18\x02 \x03(\x0b\x32\'.quantummachines.octave.api.v1.IFSource\x12<\n\x0bif_output_q\x18\x03 \x03(\x0b\x32\'.quantummachines.octave.api.v1.IFSource\x12O\n\rsynth_outputs\x18\x04 \x03(\x0b\x32\x38.quantummachines.octave.api.v1.PanelIdentity.SynthOutput\x1aR\n\x08RFOutput\x12\r\n\x05index\x18\x01 \x01(\r\x12\x37\n\x06source\x18\x02 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\x1aU\n\x0bSynthOutput\x12\r\n\x05index\x18\x01 \x01(\r\x12\x37\n\x06source\x18\x02 \x01(\x0b\x32\'.quantummachines.octave.api.v1.RFSource\")\n\x11MotherboardStatus\x12\x14\n\x0cvalid_1g_vco\x18\x01 \x01(\x08*\xf7\x01\n\x0b\x44\x65viceState\x12\x1c\n\x18\x44\x45VICE_STATE_UNSPECIFIED\x10\x00\x12 \n\x1c\x44\x45VICE_STATE_NON_OPERATIONAL\x10\x01\x12\x17\n\x13\x44\x45VICE_STATE_BOOTED\x10\x02\x12\x1b\n\x17\x44\x45VICE_STATE_CONFIGURED\x10\x03\x12\x1c\n\x18\x44\x45VICE_STATE_CLOCK_READY\x10\x04\x12\x1c\n\x18\x44\x45VICE_STATE_INITIALIZED\x10\x05\x12\x1c\n\x18\x44\x45VICE_STATE_OPERATIONAL\x10\x06\x12\x18\n\x14\x44\x45VICE_STATE_ERRORED\x10\x07*\xc4\x05\n\x0b\x44\x65viceError\x12\x1c\n\x18\x44\x45VICE_ERROR_UNSPECIFIED\x10\x00\x12\x14\n\x10\x44\x45VICE_ERROR_PLL\x10\x01\x12#\n\x1f\x44\x45VICE_ERROR_MOTHERBOARD_FW_UPD\x10\x02\x12\x1f\n\x1b\x44\x45VICE_ERROR_MODULES_DETECT\x10\x03\x12\x1d\n\x19\x44\x45VICE_ERROR_SYNTH_FW_UPD\x10\x04\x12\x1b\n\x17\x44\x45VICE_ERROR_SYNTH_COMM\x10\x05\x12\x1b\n\x17\x44\x45VICE_ERROR_START_FPGA\x10\x06\x12\x1c\n\x18\x44\x45VICE_ERROR_ATTACH_FPGA\x10\x07\x12\x19\n\x15\x44\x45VICE_ERROR_MAIN_PIC\x10\x08\x12\x1a\n\x16\x44\x45VICE_ERROR_GPIO_COMM\x10\t\x12\x1e\n\x1a\x44\x45VICE_ERROR_FPGA_OVERHEAT\x10\n\x12\x1d\n\x19\x44\x45VICE_ERROR_FPGA_HW_FAIL\x10\x0b\x12 \n\x1c\x44\x45VICE_ERROR_BAD_INPUT_CLOCK\x10\x0c\x12\x1f\n\x1b\x44\x45VICE_ERROR_SYNTH_OVERHEAT\x10\r\x12\x1f\n\x1b\x44\x45VICE_ERROR_SYNTH_BAD_READ\x10\x0e\x12&\n\"DEVICE_ERROR_IF_DOWN_CONV_OVERHEAT\x10\x0f\x12&\n\"DEVICE_ERROR_IF_DOWN_CONV_BAD_READ\x10\x10\x12&\n\"DEVICE_ERROR_RF_DOWN_CONV_OVERHEAT\x10\x11\x12&\n\"DEVICE_ERROR_RF_DOWN_CONV_BAD_READ\x10\x12\x12$\n DEVICE_ERROR_RF_UP_CONV_OVERHEAT\x10\x13\x12$\n DEVICE_ERROR_RF_UP_CONV_BAD_READ\x10\x14*\xec\x01\n\x0cOctaveModule\x12\x1d\n\x19OCTAVE_MODULE_UNSPECIFIED\x10\x00\x12 \n\x1cOCTAVE_MODULE_RF_UPCONVERTER\x10\x01\x12\"\n\x1eOCTAVE_MODULE_RF_DOWNCONVERTER\x10\x02\x12\"\n\x1eOCTAVE_MODULE_IF_DOWNCONVERTER\x10\x03\x12\x1d\n\x19OCTAVE_MODULE_SYNTHESIZER\x10\x04\x12\x1d\n\x19OCTAVE_MODULE_MOTHERBOARD\x10\x05\x12\x15\n\x11OCTAVE_MODULE_SOM\x10\x06*\x82\x01\n\x0e\x43onstantSource\x12\x1f\n\x1b\x43ONSTANT_SOURCE_UNSPECIFIED\x10\x00\x12\x1a\n\x16\x43ONSTANT_SOURCE_50_OHM\x10\x01\x12\x19\n\x15\x43ONSTANT_SOURCE_SHORT\x10\x02\x12\x18\n\x14\x43ONSTANT_SOURCE_OPEN\x10\x03\x32\x8f\x0b\n\rOctaveService\x12\x81\x01\n\x0eSubscribeState\x12\x34.quantummachines.octave.api.v1.SubscribeStateRequest\x1a\x35.quantummachines.octave.api.v1.SubscribeStateResponse\"\x00\x30\x01\x12s\n\nGetVersion\x12\x30.quantummachines.octave.api.v1.GetVersionRequest\x1a\x31.quantummachines.octave.api.v1.GetVersionResponse\"\x00\x12j\n\x07\x43ontrol\x12-.quantummachines.octave.api.v1.ControlRequest\x1a..quantummachines.octave.api.v1.ControlResponse\"\x00\x12g\n\x06Update\x12,.quantummachines.octave.api.v1.UpdateRequest\x1a-.quantummachines.octave.api.v1.UpdateResponse\"\x00\x12g\n\x06\x41quire\x12,.quantummachines.octave.api.v1.AquireRequest\x1a-.quantummachines.octave.api.v1.AquireResponse\"\x00\x12m\n\x08Identify\x12..quantummachines.octave.api.v1.IdentifyRequest\x1a/.quantummachines.octave.api.v1.IdentifyResponse\"\x00\x12j\n\x07Monitor\x12-.quantummachines.octave.api.v1.MonitorRequest\x1a..quantummachines.octave.api.v1.MonitorResponse\"\x00\x12j\n\x07\x45xplore\x12-.quantummachines.octave.api.v1.ExploreRequest\x1a..quantummachines.octave.api.v1.ExploreResponse\"\x00\x12\x61\n\x04Save\x12*.quantummachines.octave.api.v1.SaveRequest\x1a+.quantummachines.octave.api.v1.SaveResponse\"\x00\x12g\n\x06Recall\x12,.quantummachines.octave.api.v1.RecallRequest\x1a-.quantummachines.octave.api.v1.RecallResponse\"\x00\x12\x61\n\x04List\x12*.quantummachines.octave.api.v1.ListRequest\x1a+.quantummachines.octave.api.v1.ListResponse\"\x00\x12\x64\n\x05Reset\x12+.quantummachines.octave.api.v1.ResetRequest\x1a,.quantummachines.octave.api.v1.ResetResponse\"\x00\x12k\n\x06Health\x12,.quantummachines.octave.api.v1.HealthRequest\x1a-.quantummachines.octave.api.v1.HealthResponse\"\x00(\x01\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'octave.v1.api_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_DEVICESTATE']._serialized_start=12054
  _globals['_DEVICESTATE']._serialized_end=12301
  _globals['_DEVICEERROR']._serialized_start=12304
  _globals['_DEVICEERROR']._serialized_end=13012
  _globals['_OCTAVEMODULE']._serialized_start=13015
  _globals['_OCTAVEMODULE']._serialized_end=13251
  _globals['_CONSTANTSOURCE']._serialized_start=13254
  _globals['_CONSTANTSOURCE']._serialized_end=13384
  _globals['_SUBSCRIBESTATEREQUEST']._serialized_start=86
  _globals['_SUBSCRIBESTATEREQUEST']._serialized_end=109
  _globals['_SUBSCRIBESTATERESPONSE']._serialized_start=112
  _globals['_SUBSCRIBESTATERESPONSE']._serialized_end=254
  _globals['_RESETREQUEST']._serialized_start=256
  _globals['_RESETREQUEST']._serialized_end=270
  _globals['_RESETRESPONSE']._serialized_start=272
  _globals['_RESETRESPONSE']._serialized_end=330
  _globals['_HEALTHREQUEST']._serialized_start=332
  _globals['_HEALTHREQUEST']._serialized_end=402
  _globals['_HEALTHRESPONSE']._serialized_start=405
  _globals['_HEALTHRESPONSE']._serialized_end=551
  _globals['_GETVERSIONREQUEST']._serialized_start=553
  _globals['_GETVERSIONREQUEST']._serialized_end=572
  _globals['_GETVERSIONRESPONSE']._serialized_start=574
  _globals['_GETVERSIONRESPONSE']._serialized_end=611
  _globals['_MODULEREFERENCE']._serialized_start=613
  _globals['_MODULEREFERENCE']._serialized_end=704
  _globals['_SAVEREQUEST']._serialized_start=707
  _globals['_SAVEREQUEST']._serialized_end=835
  _globals['_SAVERESPONSE']._serialized_start=837
  _globals['_SAVERESPONSE']._serialized_end=885
  _globals['_RECALLREQUEST']._serialized_start=887
  _globals['_RECALLREQUEST']._serialized_end=914
  _globals['_RECALLRESPONSE']._serialized_start=916
  _globals['_RECALLRESPONSE']._serialized_end=972
  _globals['_LISTREQUEST']._serialized_start=974
  _globals['_LISTREQUEST']._serialized_end=987
  _globals['_SAVEINFO']._serialized_start=990
  _globals['_SAVEINFO']._serialized_end=1124
  _globals['_LISTRESPONSE']._serialized_start=1126
  _globals['_LISTRESPONSE']._serialized_end=1201
  _globals['_MONITORREQUEST']._serialized_start=1203
  _globals['_MONITORREQUEST']._serialized_end=1239
  _globals['_MONITORRESPONSE']._serialized_start=1242
  _globals['_MONITORRESPONSE']._serialized_end=2002
  _globals['_MONITORRESPONSE_MODULESTATUSERROR']._serialized_start=1339
  _globals['_MONITORRESPONSE_MODULESTATUSERROR']._serialized_end=1432
  _globals['_MONITORRESPONSE_MODULESTATUS']._serialized_start=1435
  _globals['_MONITORRESPONSE_MODULESTATUS']._serialized_end=1616
  _globals['_MONITORRESPONSE_OCTAVEERROR']._serialized_start=1619
  _globals['_MONITORRESPONSE_OCTAVEERROR']._serialized_end=2002
  _globals['_CONTROLREQUEST']._serialized_start=2004
  _globals['_CONTROLREQUEST']._serialized_end=2054
  _globals['_CONTROLRESPONSE']._serialized_start=2057
  _globals['_CONTROLRESPONSE']._serialized_end=2200
  _globals['_CONTROLRESPONSE_RDATADEBUG']._serialized_start=2092
  _globals['_CONTROLRESPONSE_RDATADEBUG']._serialized_end=2200
  _globals['_EXPLOREREQUEST']._serialized_start=2202
  _globals['_EXPLOREREQUEST']._serialized_end=2218
  _globals['_EXPLORERESPONSE']._serialized_start=2221
  _globals['_EXPLORERESPONSE']._serialized_end=2400
  _globals['_EXPLORERESPONSE_MODULEID']._serialized_start=2314
  _globals['_EXPLORERESPONSE_MODULEID']._serialized_end=2400
  _globals['_UPDATEREQUEST']._serialized_start=2402
  _globals['_UPDATEREQUEST']._serialized_end=2479
  _globals['_UPDATERESPONSE']._serialized_start=2481
  _globals['_UPDATERESPONSE']._serialized_end=2537
  _globals['_SINGLEUPDATE']._serialized_start=2540
  _globals['_SINGLEUPDATE']._serialized_end=2974
  _globals['_RFUPCONVUPDATE']._serialized_start=2977
  _globals['_RFUPCONVUPDATE']._serialized_end=3688
  _globals['_RFUPCONVUPDATE_IFINPUTSELECTION']._serialized_start=3394
  _globals['_RFUPCONVUPDATE_IFINPUTSELECTION']._serialized_end=3524
  _globals['_RFUPCONVUPDATE_FASTSWITCHMODE']._serialized_start=3527
  _globals['_RFUPCONVUPDATE_FASTSWITCHMODE']._serialized_end=3688
  _globals['_RFDOWNCONVUPDATE']._serialized_start=3691
  _globals['_RFDOWNCONVUPDATE']._serialized_end=4234
  _globals['_RFDOWNCONVUPDATE_RFINPUT']._serialized_start=3922
  _globals['_RFDOWNCONVUPDATE_RFINPUT']._serialized_end=4138
  _globals['_RFDOWNCONVUPDATE_LOINPUT']._serialized_start=4140
  _globals['_RFDOWNCONVUPDATE_LOINPUT']._serialized_end=4234
  _globals['_IFDOWNCONVUPDATE']._serialized_start=4237
  _globals['_IFDOWNCONVUPDATE']._serialized_end=4767
  _globals['_IFDOWNCONVUPDATE_CHANNEL']._serialized_start=4423
  _globals['_IFDOWNCONVUPDATE_CHANNEL']._serialized_end=4576
  _globals['_IFDOWNCONVUPDATE_MODE']._serialized_start=4578
  _globals['_IFDOWNCONVUPDATE_MODE']._serialized_end=4676
  _globals['_IFDOWNCONVUPDATE_COUPLING']._serialized_start=4678
  _globals['_IFDOWNCONVUPDATE_COUPLING']._serialized_end=4767
  _globals['_CLOCKUPDATE']._serialized_start=4770
  _globals['_CLOCKUPDATE']._serialized_end=5084
  _globals['_CLOCKUPDATE_MODE']._serialized_start=4985
  _globals['_CLOCKUPDATE_MODE']._serialized_end=5070
  _globals['_MOTHERBOARDUPDATE']._serialized_start=5086
  _globals['_MOTHERBOARDUPDATE']._serialized_end=5154
  _globals['_SYNTHUPDATE']._serialized_start=5157
  _globals['_SYNTHUPDATE']._serialized_end=7252
  _globals['_SYNTHUPDATE_REFERENCECLOCK']._serialized_start=5894
  _globals['_SYNTHUPDATE_REFERENCECLOCK']._serialized_end=6081
  _globals['_SYNTHUPDATE_SYNTHOUTPUT']._serialized_start=6083
  _globals['_SYNTHUPDATE_SYNTHOUTPUT']._serialized_end=6205
  _globals['_SYNTHUPDATE_STABILIZER']._serialized_start=6208
  _globals['_SYNTHUPDATE_STABILIZER']._serialized_end=6569
  _globals['_SYNTHUPDATE_REFERENCESOURCE']._serialized_start=6571
  _globals['_SYNTHUPDATE_REFERENCESOURCE']._serialized_end=6684
  _globals['_SYNTHUPDATE_SYNTHOUTPUTPOWER']._serialized_start=6687
  _globals['_SYNTHUPDATE_SYNTHOUTPUTPOWER']._serialized_end=6865
  _globals['_SYNTHUPDATE_MAINSOURCE']._serialized_start=6867
  _globals['_SYNTHUPDATE_MAINSOURCE']._serialized_end=6963
  _globals['_SYNTHUPDATE_MAINOUTPUT']._serialized_start=6965
  _globals['_SYNTHUPDATE_MAINOUTPUT']._serialized_end=7079
  _globals['_SYNTHUPDATE_SECONDARYOUTPUT']._serialized_start=7082
  _globals['_SYNTHUPDATE_SECONDARYOUTPUT']._serialized_end=7252
  _globals['_AQUIREREQUEST']._serialized_start=7254
  _globals['_AQUIREREQUEST']._serialized_end=7353
  _globals['_AQUIRERESPONSE']._serialized_start=7355
  _globals['_AQUIRERESPONSE']._serialized_end=7432
  _globals['_IDENTIFYREQUEST']._serialized_start=7434
  _globals['_IDENTIFYREQUEST']._serialized_end=7451
  _globals['_IDENTIFYRESPONSE']._serialized_start=7454
  _globals['_IDENTIFYRESPONSE']._serialized_end=7843
  _globals['_UPCONVRFOUTPUT']._serialized_start=7845
  _globals['_UPCONVRFOUTPUT']._serialized_end=7876
  _globals['_SYNTHRFOUTPUT']._serialized_start=7879
  _globals['_SYNTHRFOUTPUT']._serialized_end=8102
  _globals['_SYNTHRFOUTPUT_OUTPUTPORT']._serialized_start=7989
  _globals['_SYNTHRFOUTPUT_OUTPUTPORT']._serialized_end=8102
  _globals['_EXTERNALRFINPUT']._serialized_start=8104
  _globals['_EXTERNALRFINPUT']._serialized_end=8211
  _globals['_RFSOURCE']._serialized_start=8214
  _globals['_RFSOURCE']._serialized_end=8528
  _globals['_RFDOWNCONVIFSOURCE']._serialized_start=8531
  _globals['_RFDOWNCONVIFSOURCE']._serialized_end=8730
  _globals['_RFDOWNCONVIFSOURCE_OUTPUTPORT']._serialized_start=8651
  _globals['_RFDOWNCONVIFSOURCE_OUTPUTPORT']._serialized_end=8730
  _globals['_IFDOWNCONVIFSOURCE']._serialized_start=8732
  _globals['_IFDOWNCONVIFSOURCE']._serialized_end=8790
  _globals['_EXTERNALIFINPUT']._serialized_start=8792
  _globals['_EXTERNALIFINPUT']._serialized_end=8906
  _globals['_IFSOURCE']._serialized_start=8909
  _globals['_IFSOURCE']._serialized_end=9248
  _globals['_RFUPCONVIDENTITY']._serialized_start=9251
  _globals['_RFUPCONVIDENTITY']._serialized_end=9692
  _globals['_RFUPCONVIDENTITY_CONNECTIVITY']._serialized_start=9451
  _globals['_RFUPCONVIDENTITY_CONNECTIVITY']._serialized_end=9640
  _globals['_RFUPCONVIDENTITY_PARAMETERS']._serialized_start=9642
  _globals['_RFUPCONVIDENTITY_PARAMETERS']._serialized_end=9692
  _globals['_RFDOWNCONVIDENTITY']._serialized_start=9695
  _globals['_RFDOWNCONVIDENTITY']._serialized_end=10354
  _globals['_RFDOWNCONVIDENTITY_CONNECTIVITY']._serialized_start=9819
  _globals['_RFDOWNCONVIDENTITY_CONNECTIVITY']._serialized_end=10354
  _globals['_IFDOWNCONVIDENTITY']._serialized_start=10357
  _globals['_IFDOWNCONVIDENTITY']._serialized_end=10765
  _globals['_IFDOWNCONVIDENTITY_CONNECTIVITY']._serialized_start=10481
  _globals['_IFDOWNCONVIDENTITY_CONNECTIVITY']._serialized_end=10765
  _globals['_SYNTHIDENTITY']._serialized_start=10768
  _globals['_SYNTHIDENTITY']._serialized_end=11539
  _globals['_SYNTHIDENTITY_CONNECTIVITY']._serialized_start=10959
  _globals['_SYNTHIDENTITY_CONNECTIVITY']._serialized_end=11106
  _globals['_SYNTHIDENTITY_PARAMETERS']._serialized_start=11109
  _globals['_SYNTHIDENTITY_PARAMETERS']._serialized_end=11539
  _globals['_SYNTHIDENTITY_PARAMETERS_LOWFREQUENCYFILTER']._serialized_start=11446
  _globals['_SYNTHIDENTITY_PARAMETERS_LOWFREQUENCYFILTER']._serialized_end=11517
  _globals['_SYNTHIDENTITY_PARAMETERS_PARAMETRIZEDFILTER']._serialized_start=11519
  _globals['_SYNTHIDENTITY_PARAMETERS_PARAMETRIZEDFILTER']._serialized_end=11539
  _globals['_PANELIDENTITY']._serialized_start=11542
  _globals['_PANELIDENTITY']._serialized_end=12008
  _globals['_PANELIDENTITY_RFOUTPUT']._serialized_start=11839
  _globals['_PANELIDENTITY_RFOUTPUT']._serialized_end=11921
  _globals['_PANELIDENTITY_SYNTHOUTPUT']._serialized_start=11923
  _globals['_PANELIDENTITY_SYNTHOUTPUT']._serialized_end=12008
  _globals['_MOTHERBOARDSTATUS']._serialized_start=12010
  _globals['_MOTHERBOARDSTATUS']._serialized_end=12051
  _globals['_OCTAVESERVICE']._serialized_start=13387
  _globals['_OCTAVESERVICE']._serialized_end=14810
# @@protoc_insertion_point(module_scope)



from qm.grpc.qm.pb import inc_qua_config_pb2 as _inc_qua_config_pb2
from qm.grpc.qm.pb import general_messages_pb2 as _general_messages_pb2
from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class DigitalInputPortPolarity(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    RISING: _ClassVar[DigitalInputPortPolarity]
    FALLING: _ClassVar[DigitalInputPortPolarity]
RISING: DigitalInputPortPolarity
FALLING: DigitalInputPortPolarity

class HighQmApiRequest(_message.Message):
    __slots__ = ("config", "strongConfig", "quantumMachineId", "setCorrection", "setFrequency", "setDcOffset", "setDigitalRoute", "setIOValues", "setInputDcOffset", "setOutputDcOffset", "setOutputFilterTaps", "setDigitalInputThreshold", "setDigitalInputDeadtime", "setDigitalInputPolarity")
    class SetCorrection(_message.Message):
        __slots__ = ("correction", "qe", "mixer")
        CORRECTION_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        MIXER_FIELD_NUMBER: _ClassVar[int]
        correction: Matrix
        qe: str
        mixer: HighQmApiRequest.SetCorrectionMixerInfo
        def __init__(self, correction: _Optional[_Union[Matrix, _Mapping]] = ..., qe: _Optional[str] = ..., mixer: _Optional[_Union[HighQmApiRequest.SetCorrectionMixerInfo, _Mapping]] = ...) -> None: ...
    class SetCorrectionMixerInfo(_message.Message):
        __slots__ = ("mixer", "intermediateFrequency", "loFrequency", "frequencyNegative", "intermediateFrequencyDouble", "loFrequencyDouble")
        MIXER_FIELD_NUMBER: _ClassVar[int]
        INTERMEDIATEFREQUENCY_FIELD_NUMBER: _ClassVar[int]
        LOFREQUENCY_FIELD_NUMBER: _ClassVar[int]
        FREQUENCYNEGATIVE_FIELD_NUMBER: _ClassVar[int]
        INTERMEDIATEFREQUENCYDOUBLE_FIELD_NUMBER: _ClassVar[int]
        LOFREQUENCYDOUBLE_FIELD_NUMBER: _ClassVar[int]
        mixer: str
        intermediateFrequency: int
        loFrequency: int
        frequencyNegative: bool
        intermediateFrequencyDouble: float
        loFrequencyDouble: float
        def __init__(self, mixer: _Optional[str] = ..., intermediateFrequency: _Optional[int] = ..., loFrequency: _Optional[int] = ..., frequencyNegative: bool = ..., intermediateFrequencyDouble: _Optional[float] = ..., loFrequencyDouble: _Optional[float] = ...) -> None: ...
    class SetFrequency(_message.Message):
        __slots__ = ("value", "qe")
        VALUE_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        value: float
        qe: str
        def __init__(self, value: _Optional[float] = ..., qe: _Optional[str] = ...) -> None: ...
    class SetDcOffset(_message.Message):
        __slots__ = ("I", "Q", "qe")
        I_FIELD_NUMBER: _ClassVar[int]
        Q_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        I: float
        Q: float
        qe: HighQmApiRequest.QePort
        def __init__(self, I: _Optional[float] = ..., Q: _Optional[float] = ..., qe: _Optional[_Union[HighQmApiRequest.QePort, _Mapping]] = ...) -> None: ...
    class SetOutputDcOffset(_message.Message):
        __slots__ = ("I", "Q", "qe")
        I_FIELD_NUMBER: _ClassVar[int]
        Q_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        I: float
        Q: float
        qe: HighQmApiRequest.QePort
        def __init__(self, I: _Optional[float] = ..., Q: _Optional[float] = ..., qe: _Optional[_Union[HighQmApiRequest.QePort, _Mapping]] = ...) -> None: ...
    class SetOutputFilterTaps(_message.Message):
        __slots__ = ("filter", "qe")
        FILTER_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        filter: HighQmApiRequest.AnalogOutputPortFilter
        qe: HighQmApiRequest.QePort
        def __init__(self, filter: _Optional[_Union[HighQmApiRequest.AnalogOutputPortFilter, _Mapping]] = ..., qe: _Optional[_Union[HighQmApiRequest.QePort, _Mapping]] = ...) -> None: ...
    class AnalogOutputPortFilter(_message.Message):
        __slots__ = ("feedforward", "feedback")
        FEEDFORWARD_FIELD_NUMBER: _ClassVar[int]
        FEEDBACK_FIELD_NUMBER: _ClassVar[int]
        feedforward: _containers.RepeatedScalarFieldContainer[float]
        feedback: _containers.RepeatedScalarFieldContainer[float]
        def __init__(self, feedforward: _Optional[_Iterable[float]] = ..., feedback: _Optional[_Iterable[float]] = ...) -> None: ...
    class SetInputDcOffset(_message.Message):
        __slots__ = ("offset", "qe")
        OFFSET_FIELD_NUMBER: _ClassVar[int]
        QE_FIELD_NUMBER: _ClassVar[int]
        offset: float
        qe: HighQmApiRequest.QePort
        def __init__(self, offset: _Optional[float] = ..., qe: _Optional[_Union[HighQmApiRequest.QePort, _Mapping]] = ...) -> None: ...
    class SetDigitalRoute(_message.Message):
        __slots__ = ("value", "delay", "buffer", "tof", "smearing")
        VALUE_FIELD_NUMBER: _ClassVar[int]
        DELAY_FIELD_NUMBER: _ClassVar[int]
        BUFFER_FIELD_NUMBER: _ClassVar[int]
        TOF_FIELD_NUMBER: _ClassVar[int]
        SMEARING_FIELD_NUMBER: _ClassVar[int]
        value: int
        delay: HighQmApiRequest.QePort
        buffer: HighQmApiRequest.QePort
        tof: str
        smearing: str
        def __init__(self, value: _Optional[int] = ..., delay: _Optional[_Union[HighQmApiRequest.QePort, _Mapping]] = ..., buffer: _Optional[_Union[HighQmApiRequest.QePort, _Mapping]] = ..., tof: _Optional[str] = ..., smearing: _Optional[str] = ...) -> None: ...
    class SetIOValues(_message.Message):
        __slots__ = ("all", "ioValueSetData")
        ALL_FIELD_NUMBER: _ClassVar[int]
        IOVALUESETDATA_FIELD_NUMBER: _ClassVar[int]
        all: bool
        ioValueSetData: _containers.RepeatedCompositeFieldContainer[HighQmApiRequest.IOValueSetData]
        def __init__(self, all: bool = ..., ioValueSetData: _Optional[_Iterable[_Union[HighQmApiRequest.IOValueSetData, _Mapping]]] = ...) -> None: ...
    class QePort(_message.Message):
        __slots__ = ("qe", "port")
        QE_FIELD_NUMBER: _ClassVar[int]
        PORT_FIELD_NUMBER: _ClassVar[int]
        qe: str
        port: str
        def __init__(self, qe: _Optional[str] = ..., port: _Optional[str] = ...) -> None: ...
    class IOValueSetData(_message.Message):
        __slots__ = ("io_number", "intValue", "doubleValue", "booleanValue")
        IO_NUMBER_FIELD_NUMBER: _ClassVar[int]
        INTVALUE_FIELD_NUMBER: _ClassVar[int]
        DOUBLEVALUE_FIELD_NUMBER: _ClassVar[int]
        BOOLEANVALUE_FIELD_NUMBER: _ClassVar[int]
        io_number: int
        intValue: int
        doubleValue: float
        booleanValue: bool
        def __init__(self, io_number: _Optional[int] = ..., intValue: _Optional[int] = ..., doubleValue: _Optional[float] = ..., booleanValue: bool = ...) -> None: ...
    class SetDigitalInputThreshold(_message.Message):
        __slots__ = ("digitalPort", "threshold")
        DIGITALPORT_FIELD_NUMBER: _ClassVar[int]
        THRESHOLD_FIELD_NUMBER: _ClassVar[int]
        digitalPort: DigitalInputPort
        threshold: float
        def __init__(self, digitalPort: _Optional[_Union[DigitalInputPort, _Mapping]] = ..., threshold: _Optional[float] = ...) -> None: ...
    class SetDigitalInputPolarity(_message.Message):
        __slots__ = ("digitalPort", "polarity")
        DIGITALPORT_FIELD_NUMBER: _ClassVar[int]
        POLARITY_FIELD_NUMBER: _ClassVar[int]
        digitalPort: DigitalInputPort
        polarity: DigitalInputPortPolarity
        def __init__(self, digitalPort: _Optional[_Union[DigitalInputPort, _Mapping]] = ..., polarity: _Optional[_Union[DigitalInputPortPolarity, str]] = ...) -> None: ...
    class SetDigitalInputDeadtime(_message.Message):
        __slots__ = ("digitalPort", "deadtime")
        DIGITALPORT_FIELD_NUMBER: _ClassVar[int]
        DEADTIME_FIELD_NUMBER: _ClassVar[int]
        digitalPort: DigitalInputPort
        deadtime: int
        def __init__(self, digitalPort: _Optional[_Union[DigitalInputPort, _Mapping]] = ..., deadtime: _Optional[int] = ...) -> None: ...
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    STRONGCONFIG_FIELD_NUMBER: _ClassVar[int]
    QUANTUMMACHINEID_FIELD_NUMBER: _ClassVar[int]
    SETCORRECTION_FIELD_NUMBER: _ClassVar[int]
    SETFREQUENCY_FIELD_NUMBER: _ClassVar[int]
    SETDCOFFSET_FIELD_NUMBER: _ClassVar[int]
    SETDIGITALROUTE_FIELD_NUMBER: _ClassVar[int]
    SETIOVALUES_FIELD_NUMBER: _ClassVar[int]
    SETINPUTDCOFFSET_FIELD_NUMBER: _ClassVar[int]
    SETOUTPUTDCOFFSET_FIELD_NUMBER: _ClassVar[int]
    SETOUTPUTFILTERTAPS_FIELD_NUMBER: _ClassVar[int]
    SETDIGITALINPUTTHRESHOLD_FIELD_NUMBER: _ClassVar[int]
    SETDIGITALINPUTDEADTIME_FIELD_NUMBER: _ClassVar[int]
    SETDIGITALINPUTPOLARITY_FIELD_NUMBER: _ClassVar[int]
    config: QmConfig
    strongConfig: _inc_qua_config_pb2.QuaConfig
    quantumMachineId: str
    setCorrection: HighQmApiRequest.SetCorrection
    setFrequency: HighQmApiRequest.SetFrequency
    setDcOffset: HighQmApiRequest.SetDcOffset
    setDigitalRoute: HighQmApiRequest.SetDigitalRoute
    setIOValues: HighQmApiRequest.SetIOValues
    setInputDcOffset: HighQmApiRequest.SetInputDcOffset
    setOutputDcOffset: HighQmApiRequest.SetOutputDcOffset
    setOutputFilterTaps: HighQmApiRequest.SetOutputFilterTaps
    setDigitalInputThreshold: HighQmApiRequest.SetDigitalInputThreshold
    setDigitalInputDeadtime: HighQmApiRequest.SetDigitalInputDeadtime
    setDigitalInputPolarity: HighQmApiRequest.SetDigitalInputPolarity
    def __init__(self, config: _Optional[_Union[QmConfig, _Mapping]] = ..., strongConfig: _Optional[_Union[_inc_qua_config_pb2.QuaConfig, _Mapping]] = ..., quantumMachineId: _Optional[str] = ..., setCorrection: _Optional[_Union[HighQmApiRequest.SetCorrection, _Mapping]] = ..., setFrequency: _Optional[_Union[HighQmApiRequest.SetFrequency, _Mapping]] = ..., setDcOffset: _Optional[_Union[HighQmApiRequest.SetDcOffset, _Mapping]] = ..., setDigitalRoute: _Optional[_Union[HighQmApiRequest.SetDigitalRoute, _Mapping]] = ..., setIOValues: _Optional[_Union[HighQmApiRequest.SetIOValues, _Mapping]] = ..., setInputDcOffset: _Optional[_Union[HighQmApiRequest.SetInputDcOffset, _Mapping]] = ..., setOutputDcOffset: _Optional[_Union[HighQmApiRequest.SetOutputDcOffset, _Mapping]] = ..., setOutputFilterTaps: _Optional[_Union[HighQmApiRequest.SetOutputFilterTaps, _Mapping]] = ..., setDigitalInputThreshold: _Optional[_Union[HighQmApiRequest.SetDigitalInputThreshold, _Mapping]] = ..., setDigitalInputDeadtime: _Optional[_Union[HighQmApiRequest.SetDigitalInputDeadtime, _Mapping]] = ..., setDigitalInputPolarity: _Optional[_Union[HighQmApiRequest.SetDigitalInputPolarity, _Mapping]] = ...) -> None: ...

class DigitalInputPort(_message.Message):
    __slots__ = ("controllerName", "portNumber", "fem_Number")
    CONTROLLERNAME_FIELD_NUMBER: _ClassVar[int]
    PORTNUMBER_FIELD_NUMBER: _ClassVar[int]
    FEM_NUMBER_FIELD_NUMBER: _ClassVar[int]
    controllerName: str
    portNumber: int
    fem_Number: int
    def __init__(self, controllerName: _Optional[str] = ..., portNumber: _Optional[int] = ..., fem_Number: _Optional[int] = ...) -> None: ...

class HighQmApiResponse(_message.Message):
    __slots__ = ("ok", "errors")
    OK_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    ok: bool
    errors: _containers.RepeatedCompositeFieldContainer[_general_messages_pb2.ErrorMessage]
    def __init__(self, ok: bool = ..., errors: _Optional[_Iterable[_Union[_general_messages_pb2.ErrorMessage, _Mapping]]] = ...) -> None: ...

class QmConfig(_message.Message):
    __slots__ = ("version", "root")
    VERSION_FIELD_NUMBER: _ClassVar[int]
    ROOT_FIELD_NUMBER: _ClassVar[int]
    version: int
    root: _struct_pb2.Struct
    def __init__(self, version: _Optional[int] = ..., root: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...) -> None: ...

class Matrix(_message.Message):
    __slots__ = ("v00", "v01", "v10", "v11")
    V00_FIELD_NUMBER: _ClassVar[int]
    V01_FIELD_NUMBER: _ClassVar[int]
    V10_FIELD_NUMBER: _ClassVar[int]
    V11_FIELD_NUMBER: _ClassVar[int]
    v00: float
    v01: float
    v10: float
    v11: float
    def __init__(self, v00: _Optional[float] = ..., v01: _Optional[float] = ..., v10: _Optional[float] = ..., v11: _Optional[float] = ...) -> None: ...

class PortReference(_message.Message):
    __slots__ = ("controller", "number")
    CONTROLLER_FIELD_NUMBER: _ClassVar[int]
    NUMBER_FIELD_NUMBER: _ClassVar[int]
    controller: str
    number: int
    def __init__(self, controller: _Optional[str] = ..., number: _Optional[int] = ...) -> None: ...



from typing import Dict, MutableMapping, cast

from qm.grpc.qm.pb import inc_qua_config_pb2
from qm.utils.protobuf_utils import which_one_of
from qm.api.models.server_details import ConnectionDetails
from qm.api.v2.job_api.element_api import ElementApi, JobElement
from qm.api.v2.job_api.element_input_api import InputConfigType, create_element_input_class
from qm.api.v2.job_api.element_port_api import MwOutputApi, AnalogOutputApi, DigitalInputApi


class JobElementNotFound(KeyError):
    def __init__(self, key: str):
        self._key = key

    def __str__(self) -> str:
        return f"Element with the key {self._key} was not found."


class JobElementsDB(Dict[str, JobElement]):
    def __missing__(self, key: str) -> None:
        raise JobElementNotFound(key)

    @classmethod
    def init_from_data(
        cls,
        elements: MutableMapping[str, inc_qua_config_pb2.QuaConfig.ElementDec],
        connection_details: ConnectionDetails,
        job_id: str,
    ) -> "JobElementsDB":
        res = {
            element_name: cls._create_single_element(element_name, element_config, connection_details, job_id)
            for element_name, element_config in elements.items()
        }
        return cls(res)

    @staticmethod
    def _create_single_element(
        element_name: str,
        element_config: inc_qua_config_pb2.QuaConfig.ElementDec,
        connection_details: ConnectionDetails,
        job_id: str,
    ) -> JobElement:
        input_config = cast(InputConfigType, which_one_of(element_config, "element_inputs_one_of")[1])
        input_api_class = create_element_input_class(input_config)

        mw_output_api = None
        outputs = which_one_of(element_config, "element_outputs_one_of")[1]
        if isinstance(outputs, inc_qua_config_pb2.QuaConfig.MultipleOutputs):
            output_names = list(outputs.port_references)
        elif isinstance(outputs, inc_qua_config_pb2.QuaConfig.MicrowaveOutputPortReference):
            output_names = list(element_config.outputs)
            mw_output_api = MwOutputApi(connection_details, job_id, element_name)
        else:
            output_names = list(element_config.outputs)

        outputs_apis = {
            output_name: AnalogOutputApi(connection_details, job_id, element_name, output_name)
            for output_name in output_names
        }
        digital_inputs_apis = {
            input_name: DigitalInputApi(connection_details, job_id, element_name, input_name)
            for input_name in element_config.digitalInputs
        }
        return JobElement(
            api=ElementApi(connection_details, job_id, element_name),
            input_api=input_api_class(connection_details, job_id, element_name),
            outputs_apis=outputs_apis,
            digital_inputs_apis=digital_inputs_apis,
            microwave_output_api=mw_output_api,
        )



from collections import Counter
from typing import Dict, Type, Optional, cast

from qm.type_hinting import NumberT
from qm.grpc.qm.pb import inc_qua_pb2
from qm.exceptions import QmQuaException
from qm.qua._dsl.scope_functions import stream_processing
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._dsl.variable_handling import save, declare, _get_save_statement
from qm.qua._scope_management._core_scopes import _LoopScope, _PythonNativeScope
from qm.qua._dsl.streams.output_streams.declare_output_stream import declare_output_stream
from qm.qua._dsl.stream_processing.stream_processing import ResultStream, ResultStreamSource
from qm.qua._dsl.stream_processing.direct_stream_processing_interface import DirectStreamSourceInterface

STREAM_NAME_SEPARATOR = "__"


class DirectStreamSource(DirectStreamSourceInterface[NumberT]):
    @classmethod
    def get_existing_direct_streams(
        cls,
        stream_name: str,
        var_type: Type[NumberT],
        init_value: Optional[NumberT],
        auto_buffer: bool = False,
        average_axes: Optional[list[str]] = None,
    ) -> "DirectStreamSource[NumberT]":
        """Get existing streams associated with this DirectStreamSource."""
        matching_stream = [
            stream
            for stream in scopes_manager.program_scope.auto_processing_streams
            if stream.stream_name == stream_name
        ]

        # add axis to average stream if exist
        # doing it here as native context is duplicating the get call but not the init call
        if average_axes:
            for scope in scopes_manager.scope_stack:
                if isinstance(scope, _LoopScope):
                    if scope.name in average_axes:
                        scope.add_averaged_stream(stream_name)

        if len(matching_stream) == 1:
            return cast("DirectStreamSource[NumberT]", matching_stream[0])
        elif len(matching_stream) > 1:
            raise QmQuaException(f"Multiple DirectStreamSource found with the same stream name '{stream_name}'.")
        else:
            stream = cls(
                stream_name=stream_name,
                var_type=var_type,
                init_value=init_value,
                auto_buffer=auto_buffer,
                average_axes=average_axes,
            )

            # add to auto stream processing
            scopes_manager.program_scope.add_auto_processing_stream(stream)
            return stream

    def __init__(
        self,
        stream_name: str,
        var_type: Type[NumberT],
        init_value: Optional[NumberT],
        auto_buffer: bool = False,
        average_axes: Optional[list[str]] = None,
    ) -> None:

        # declare the variable
        if init_value is not None:
            qua_var = declare(var_type, init_value)
        else:
            qua_var = declare(var_type)

        super().__init__(qua_var.unwrapped_scalar.name, var_type, init_value=init_value)  # type: ignore[arg-type]
        self._streams: Dict[str, ResultStreamSource] = {}
        self._stream_name = stream_name

        self._auto_buffer = auto_buffer
        self._average_axes = average_axes if average_axes is not None else []
        self._save_calls: Counter[str] = Counter()
        self._calculate_buffers()

    @property
    def stream_name(self) -> str:
        return self._stream_name

    @property
    def stream_full_name(self) -> str:
        """Get the stream name, potentially modified by native loop contexts."""
        stream_full_name = self.stream_name
        # Append native loop contexts to the stream name
        native_contexts_values = [
            scope.current_value_name
            for scope in scopes_manager.scope_stack
            if isinstance(scope, _PythonNativeScope) and scope.name not in self._average_axes
        ]
        if len(native_contexts_values) > 0:
            stream_full_name = STREAM_NAME_SEPARATOR.join([stream_full_name] + native_contexts_values)
        return stream_full_name

    def get_stream(self) -> ResultStreamSource:
        """
        Get or create the stream associated with DirectStreamSource for current scope.
        """
        stream_full_name = self.stream_full_name
        if stream_full_name not in self._streams:
            self._streams[stream_full_name] = declare_output_stream()
        return self._streams[stream_full_name]

    def stream_processing(self, current_processed_streams: set[str]) -> None:
        """
        Process and save the stream associated with DirectStreamSource.

        Mutates current_processed_streams in place, adding the names of streams processed.

        Args:
            current_processed_streams: Set of stream names already processed in the current context.
        """
        with stream_processing():
            for stream_full_name, stream in self._streams.items():
                if stream_full_name not in current_processed_streams:
                    # Mark the stream as processed
                    current_processed_streams.add(stream_full_name)
                    save_object: ResultStream = stream
                    # Apply buffering and averaging if auto_buffer is enabled
                    if self._auto_buffer:
                        if self._buffers:
                            buffers = list(self._buffers)
                            buffers[-1] *= self._save_calls[stream_full_name]

                            if self.skip_first_buffer:
                                buffers = buffers[1:]

                            if len(buffers) > 0:
                                save_object = save_object.buffer(*buffers)

                        # Apply averaging if average_axes is specified
                        if len(self._average_axes) > 0:
                            save_object = save_object.average()

                    if self._use_save_all:
                        save_object.save_all(stream_full_name)
                    else:
                        save_object.save(stream_full_name)

    def save(self) -> None:
        """Save the variable to its associated stream."""
        self._save_calls[self.stream_full_name] += 1
        save(self, self.get_stream())

    def get_save_statement(self) -> inc_qua_pb2.QuaProgram.AnyStatement:
        """Get the save statement for saving the variable to its associated stream."""
        self._save_calls[self.stream_full_name] += 1
        return _get_save_statement(self, self.get_stream())

    def _calculate_buffers(self) -> None:
        """
        Calculate buffer sizes and average dimensions based on the current scope stack.
        """
        buffers: list[int] = []
        if self._auto_buffer:
            axes_to_average = self._average_axes
            for scope in scopes_manager.scope_stack:
                if isinstance(scope, _PythonNativeScope):
                    # skip native scopes for buffering as they are handled in the stream name
                    if scope.name in axes_to_average:
                        raise QmQuaException(f"Auto buffering cannot average over native iterator '{scope.name}'.")
                    continue

                elif isinstance(scope, _LoopScope):
                    scope_name = scope.name
                    scope_size = scope.size
                    if scope_name is None or scope_size is None:
                        raise QmQuaException(
                            "Auto buffering requires all enclosing loops to have a name and size defined."
                        )
                    # determine if we need to average over this axis
                    if scope_name not in axes_to_average:
                        buffers.append(scope_size)
                    elif buffers:
                        raise QmQuaException(
                            f"Cannot average over '{scope_name}' because non-averaged loop scopes "
                            f"exist before it. All qua iterables averaged axes must be outermost in the loop nesting."
                        )

        self.skip_first_buffer = len(buffers) > 0 and len(self._average_axes) == 0
        self._use_save_all = len(self._average_axes) == 0
        self._buffers = buffers


def declare_with_stream(
    t: type[NumberT],
    stream_name: str,
    value: Optional[NumberT] = None,
    auto_buffer: bool = True,
    average_axes: Optional[list[str]] = None,
) -> DirectStreamSource[NumberT]:
    """
    Declare a QUA variable that is automatically saved to a result stream.

    The returned object can be used anywhere the declared QUA variable can be
    used, including assignments and measurement processes. During
    [stream_processing][qm.qua.stream_processing], the saved values are emitted
    to ``stream_name``.

    When ``auto_buffer=True``, buffering is derived from the enclosing named
    QUA loops. Native iterables are not buffered; their current values are
    appended to the saved stream name instead. When no averaging is requested,
    the outermost buffered QUA axis is left unbuffered so live plotting can
    consume the stream progressively, and the generated stream-processing step
    stores the result with ``save_all()``. When ``average_axes`` is provided,
    the named axes are reduced during generated stream processing and the final
    averaged result is stored with ``save()``.

    Args:
        t: Type of the declared variable, for example ``fixed``, ``int``, or
            ``bool``.
        stream_name: Base name of the result stream.
        value: Optional initial value for the declared variable.
        auto_buffer: If ``True``, derive buffering from the enclosing named
            QUA loops. When no averaging is requested, the first buffered QUA
            axis is left unbuffered for live plotting. Defaults to ``True``.
        average_axes: Names of QUA loop axes to average when
            ``auto_buffer=True``. When provided, the generated stream
            processing stores the reduced result with ``save()``; otherwise it
            stores the stream with ``save_all()``. Defaults to ``None``.

    Returns:
        A stream-backed QUA variable.

    Raises:
        QmQuaException: If ``average_axes`` is provided while
            ``auto_buffer`` is ``False``, if ``average_axes`` refers to a
            native iterator, if non-averaged QUA loop scopes appear before an
            averaged axis, or if auto-buffering is used inside loops without
            names or known sizes.

    Example:
        ```python
        shots = QuaIterableRange("shots", 100)

        with program() as prog:
            for shot in shots:
                I = declare_with_stream(fixed, "I", auto_buffer=True)
                Q = declare_with_stream(fixed, "Q", auto_buffer=True)
                measure(
                    "readout",
                    "rr",
                    demod.full("cos", I),
                    demod.full("sin", Q),
                )
        ```
    """
    if average_axes is not None and not auto_buffer:
        raise QmQuaException("average_axes can only be used when auto_buffer is set to True.")

    combined_stream = DirectStreamSource.get_existing_direct_streams(
        stream_name=stream_name,
        var_type=t,
        init_value=value,
        auto_buffer=auto_buffer,
        average_axes=average_axes,
    )

    return combined_stream
